@@ -1,7 +1,7 @@
 /*
  *  main_beos.cpp - Startup code for BeOS
  *
- *  Basilisk II (C) 1997-2002 Christian Bauer
+ *  Basilisk II (C) 1997-2008 Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -271,7 +271,7 @@ void BasiliskII::StartEmulator(void)
 	}
 
 	// Initialize everything
-	if (!InitAll()) {
+	if (!InitAll(NULL)) {
 		PostMessage(B_QUIT_REQUESTED);
 		return;
 	}
