@@ -2,9 +2,9 @@
  *	PrefsEditor.m - GUI stuff for Basilisk II preferences
  *					(which is a text file in the user's home directory)
  *
- *	$Id: PrefsEditor.mm,v 1.10 2003/04/02 02:19:53 nigel Exp $
+ *	$Id: PrefsEditor.mm,v 1.21 2009/11/06 21:33:03 nigel Exp $
  *
- *  Basilisk II (C) 1997-2003 Christian Bauer
+ *  Basilisk II (C) 1997-2008 Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -105,8 +105,6 @@
 
 @end
 
-@implementation PrefsEditor
-
 #import <AppKit/NSImage.h>		// For [NSBundle pathForImageResource:] proto
 
 #include <string>
@@ -116,12 +114,16 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 #import "sysdeps.h"				// Types used in Basilisk C++ code
 #import "video_macosx.h"		// some items that we edit here
 #import "misc_macosx.h"			// WarningSheet() prototype
+#import "main_macosx.h"			// ChoiceAlert() prototype
+
 
 #import <prefs.h>
 
 #define DEBUG 0
 #import <debug.h>
 
+@implementation PrefsEditor
+
 - (PrefsEditor *) init
 {
 	self = [super init];
@@ -129,7 +131,7 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 	edited = NO;
 
 	devs = @"/dev";
-	home = NSHomeDirectory();
+	home = [NSHomeDirectory() retain];
 	volsDS = [TableDS new];
 	SCSIds = [TableDS new];
 
@@ -138,10 +140,9 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 		NSLog (@"%s - Can't create NSImageCell?", __PRETTY_FUNCTION__);
 
 	blank  = [NSImage new];
-	locked = [NSImage alloc];
-	if ( [locked initWithContentsOfFile:
-				 [[NSBundle mainBundle]
-						pathForImageResource: @"nowrite.icns"]] == nil )
+	locked = [[NSImage alloc] initWithContentsOfFile:
+				[[NSBundle mainBundle] pathForImageResource: @"nowrite.icns"]];
+	if (locked == nil )
 		NSLog(@"%s - Couldn't open write protection image", __PRETTY_FUNCTION__);
 
 	return self;
@@ -149,11 +150,12 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 
 - (void) dealloc
 {
-	[volsDS   dealloc];
-	[SCSIds   dealloc];
-	[lockCell dealloc];
-	[locked   dealloc];
-	[blank    dealloc];
+	[home     release];
+	[volsDS   release];
+	[SCSIds   release];
+	[lockCell release];
+	[blank    release];
+	[locked   release];
 	[super    dealloc];
 }
 
@@ -196,7 +198,7 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 	{
 		[volsDS addObject: (NSObject *) locked
 				 withPath: [oP filename] ];
-		PrefsAddString("disk", [[oP filename] cString]);
+		PrefsAddString("disk", [[oP filename] UTF8String]);
 		[diskImages reloadData];
 		edited = YES;
 	}
@@ -217,7 +219,7 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 							types:nil] == NSOKButton )
 	{
 		[extFS setStringValue: [oP directory] ];
-		PrefsReplaceString("extfs", [[oP directory] cString]);
+		PrefsReplaceString("extfs", [[oP directory] UTF8String]);
 		edited = YES;
 	}
 }
@@ -234,7 +236,7 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 							types:nil] == NSOKButton )
 	{
 		[prefsFile setStringValue: [oP filename] ];
-		UserPrefsPath = [[oP filename] cString];
+		UserPrefsPath = [[oP filename] UTF8String];
 	}
 }
 
@@ -250,7 +252,7 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 							types:nil] == NSOKButton )
 	{
 		[ROMfile setStringValue: [oP filename] ];
-		PrefsReplaceString("rom", [[oP filename] cString]);
+		PrefsReplaceString("rom", [[oP filename] UTF8String]);
 		edited = YES;
 	}
 }
@@ -259,8 +261,12 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 
 - (IBAction) ChangeBootFrom: (NSMatrix *)sender
 {
-	if ( [sender selectedCell] == bootFromCD )
+	if ( [sender selectedCell] == (id)bootFromCD )
+	{
+		[disableCD setState: NSOffState];
+
 		PrefsReplaceInt32("bootdriver", CDROMRefNum);
+	}
 	else
 		PrefsReplaceInt32("bootdriver", 0);
 	edited = YES;
@@ -274,7 +280,14 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 
 - (IBAction) ChangeDisableCD: (NSButton *)sender
 {
-	PrefsReplaceBool("nocdrom", [disableCD state]);
+	int disabled = [disableCD state];
+
+	PrefsReplaceBool("nocdrom", disabled);
+	if ( disabled )
+	{
+		[bootFromAny setState: NSOnState];
+		[bootFromCD setState: ![disableCD state]];
+	}
 	edited = YES;
 }
 
@@ -295,6 +308,17 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 	edited = YES;
 }
 
+- (IBAction) ChangeKeyboard: (NSPopUpButton *)sender
+{
+	// Deselest current item
+	int  current = [keyboard indexOfItemWithTag: PrefsFindInt32("keyboardtype")];
+	if ( current != -1 )
+		[[keyboard itemAtIndex: current] setState: FALSE];
+
+	PrefsReplaceInt32("keyboardtype", [[sender selectedItem] tag]);
+	edited = YES;
+}
+
 - (IBAction) ChangeModel: (NSMatrix *)sender
 {
 	PrefsReplaceInt32("modelid", [[sender selectedCell] tag]);
@@ -312,7 +336,7 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 #else
 	if ( newbpp != 32 )
 		WarningSheet(@"Sorry - In windowed mode, depth must be 32", panel);
-	return 32
+	return 32;
 #endif
 }
 
@@ -423,11 +447,12 @@ extern string UserPrefsPath;	// from prefs_unix.cpp
 	if ( [sP runModalForDirectory:NSHomeDirectory() file:@"basilisk-II.vol"] == NSOKButton )
 	{
 		char		cmd[1024];
-		const char	*filename = [[sP filename] cString];
+		const char	*filename = [[sP filename] UTF8String];
 		int			retVal,
 					size = [newVolumeSize intValue];
 
 		sprintf(cmd, "dd if=/dev/zero \"of=%s\" bs=1024k count=%d", filename, size);
+
 		retVal = system(cmd);
 		if (retVal != 0)
 		{
@@ -460,15 +485,17 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 
 - (IBAction) DeleteVolume: (id)sender
 {
-//	const char *path = [self RemoveVolumeEntry];
-//	if ( unlink(path) == -1 )
 	NSString	*Path = [self RemoveVolumeEntry];
 
+	if ( ! Path )
+		return;
+
 	if ( ! [[NSFileManager defaultManager] removeFileAtPath: Path
 													handler: self] )
 	{
 		WarningSheet(@"Unable to delete volume", panel);
-//		NSLog(@"%s unlink(%s) failed - %s", __PRETTY_FUNCTION__, path, strerror(errno));
+		NSLog(@"%s unlink(%s) failed - %s", __PRETTY_FUNCTION__,
+									[Path cString], strerror(errno));
 	}
 }
 
@@ -503,7 +530,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 {
 	NSString	*path = [etherNet stringValue];
 
-	PrefsReplaceString("ether", [path cString]);
+	PrefsReplaceString("ether", [path UTF8String]);
 	edited = YES;
 }
 
@@ -511,7 +538,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 {
 	NSString	*path = [extFS stringValue];
 
-	PrefsReplaceString("extfs", [path cString]);
+	PrefsReplaceString("extfs", [path UTF8String]);
 	edited = YES;
 }
 
@@ -528,7 +555,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 {
 	NSString	*path = [modem stringValue];
 
-	PrefsReplaceString("seriala", [path cString]);
+	PrefsReplaceString("seriala", [path UTF8String]);
 	edited = YES;
 }
 
@@ -547,7 +574,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 {
 	NSString	*path = [printer stringValue];
 
-	PrefsReplaceString("serialb", [path cString]);
+	PrefsReplaceString("serialb", [path UTF8String]);
 	edited = YES;
 }
 
@@ -555,7 +582,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 {
 	NSString	*path = [ROMfile stringValue];
 
-	PrefsReplaceString("rom", [path cString]);
+	PrefsReplaceString("rom", [path UTF8String]);
 }
 
 - (IBAction) RemoveSCSI: (id)sender
@@ -568,10 +595,10 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 		NSLog (@"%s - [SCSIds deleteRow: %d] failed", __PRETTY_FUNCTION__, row);
 	[SCSIdisks reloadData];
 	sprintf(pref, "scsi%d", SCSIid);
-	PrefsRemoveItem(pref,0);
+	//PrefsRemoveItem(pref,0);
+	PrefsRemoveItem(pref, 1);
 }
 
-//- (const char *) RemoveVolumeEntry
 - (NSString *) RemoveVolumeEntry
 {
 	int		row = [diskImages selectedRow];
@@ -579,10 +606,17 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 	if ( row != -1 )
 	{
 		NSString	*Path = [volsDS pathAtRow: row];
-		const char	*path = [Path cString],
+		const char	*path = [Path UTF8String],
 					*str;
 		int			tmp = 0;
 
+		NSString	*prompt = [NSString stringWithFormat: @"%s\n%s",
+							   "Are you sure you want to delete the file",
+							   path];
+
+		if ( ! ChoiceAlert([prompt cString], "Delete", "Cancel") )
+			return NULL;
+
 		while ( (str = PrefsFindString("disk", tmp) ) != NULL )
 		{
 			if ( strcmp(str, path) == 0 )
@@ -628,7 +662,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 
 	PrefsExit();				// Purge all the old pref values
 
-	PrefsInit(argc, argv);
+	PrefsInit(NULL, argc, argv);
 	AddPrefsDefaults();
 	AddPlatformPrefsDefaults();	// and only create basic ones
 
@@ -645,7 +679,7 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 	char	*argv[2];
 
 	argv[0] = "--prefs",
-	argv[1] = (char *) [[prefsFile stringValue] cString];
+	argv[1] = (char *) [[prefsFile stringValue] UTF8String];
 
 	[self loadPrefs: argc
 			   args: argv];
@@ -658,12 +692,12 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 }
 
 - (void) setStringOf: (NSTextField *) field
-			fromPref: (const char *)  prefName
+				fromPref: (const char *)  prefName
 {
 	const char	*value = PrefsFindString(prefName, 0);
 
 	if ( value )
-		[field setStringValue: [NSString stringWithCString: value] ];
+		[field setStringValue: [NSString stringWithUTF8String: value] ];
 }
 
 - (IBAction) SavePrefs: (id)sender
@@ -676,21 +710,21 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 {
 	NSTableColumn	*locks;
 	const char		*str;
-	int				cpu, tmp;
+	int				cpu, tmp, val;
 
 
 	// Set simple single field items
 
-	tmp = PrefsFindInt32("frameskip");
-	[delay setIntValue: tmp];
-	if ( tmp )
-		[frequency	setFloatValue:	60.0 / tmp];
+	val = PrefsFindInt32("frameskip");
+	[delay setIntValue: val];
+	if ( val )
+		[frequency	setFloatValue:	60.0 / val];
 	else
 		[frequency	setFloatValue:	60.0];
 
-	tmp = PrefsFindInt32("ramsize");
-	[bytes	setIntValue:   tmp];
-	[MB		setFloatValue: tmp / (1024.0 * 1024.0)];
+	val = PrefsFindInt32("ramsize");
+	[bytes	setIntValue:   val];
+	[MB		setFloatValue: val / (1024.0 * 1024.0)];
 
 	[disableCD		setState:	PrefsFindBool("nocdrom")];
 	[disableSound	setState:	PrefsFindBool("nosound")];
@@ -700,9 +734,9 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 	[self setStringOf: extFS	 fromPref: "extfs"	];
 	[self setStringOf: modem	 fromPref: "seriala"];
 	[self setStringOf: printer	 fromPref: "serialb"];
-    [self setStringOf: ROMfile	 fromPref: "rom"	];
+    [self setStringOf: ROMfile	 fromPref: "rom"];
 
-	[prefsFile setStringValue: [NSString stringWithCString: UserPrefsPath.c_str()] ];
+	[prefsFile setStringValue: [NSString stringWithUTF8String: UserPrefsPath.c_str()] ];
 
 
 	parse_screen_prefs(PrefsFindString("screen"));
@@ -722,16 +756,16 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 
 	// Radio button groups:
 
-	tmp = PrefsFindInt32("bootdriver");
-	[bootFromAny setState: tmp != CDROMRefNum];
-	[bootFromCD  setState: tmp == CDROMRefNum];
+	val = PrefsFindInt32("bootdriver");
+	[bootFromAny setState: val != CDROMRefNum];
+	[bootFromCD  setState: val == CDROMRefNum];
 
 	cpu = PrefsFindInt32("cpu");
-	tmp = PrefsFindInt32("modelid");
+	val = PrefsFindInt32("modelid");
 
 #if REAL_ADDRESSING || DIRECT_ADDRESSING
 	puts("Current memory model does not support 24bit addressing");
-	if ( tmp == [classic tag] )
+	if ( val == [classic tag] )
 	{
 		// Window already created by NIB file, just display
 		[panel makeKeyAndOrderFront:self];
@@ -739,8 +773,8 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 						@"Disabling Mac Classic emulation", nil, panel);
 		cpu = [CPU68030 tag];
 		PrefsReplaceInt32("cpu", cpu);
-		tmp = [IIci tag];
-		PrefsReplaceInt32("modelid", tmp);
+		val = [IIci tag];
+		PrefsReplaceInt32("modelid", val);
 	}
 
 	puts("Disabling 68000 & Mac Classic buttons");
@@ -753,13 +787,24 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 	[CPU68030   setState: [CPU68030  tag] == cpu];
 	[CPU68040   setState: [CPU68040  tag] == cpu];
 
-	[classic	setState: [classic	 tag] == tmp];
-	[IIci		setState: [IIci		 tag] == tmp];
-	[quadra900	setState: [quadra900 tag] == tmp];
+	[classic	setState: [classic	 tag] == val];
+	[IIci		setState: [IIci		 tag] == val];
+	[quadra900	setState: [quadra900 tag] == val];
 
 
 	// Lists of thingies:
 
+	val = PrefsFindInt32("keyboardtype");
+	tmp = [keyboard indexOfItemWithTag: val];
+	if ( tmp != -1 )
+		[keyboard selectItemAtIndex: tmp];
+	for ( tmp = 0; tmp < [keyboard numberOfItems]; ++tmp )
+	{
+		NSMenuItem	*type = [keyboard itemAtIndex: tmp];
+		[type setState: [type tag] == val];
+	}
+
+
 	for ( tmp = 0; tmp < 7; ++tmp)
 	{
 		char pref[6];
@@ -784,10 +829,10 @@ shouldProceedAfterError: (NSDictionary *) errorDict
 	{
 		if ( *str == '*' )
 			[volsDS addObject: (NSObject *) locked
-					 withPath: [NSString stringWithCString: str+1]];
+					 withPath: [NSString stringWithUTF8String: str+1]];
 		else
 			[volsDS addObject: (NSObject *) blank
-					 withPath: [NSString stringWithCString: str]];
+					 withPath: [NSString stringWithUTF8String: str]];
 	}
 
 	[diskImages setDataSource: volsDS];
