@@ -1,10 +1,10 @@
 /*
- *  $Id: video_macosx.mm,v 1.10 2003/03/26 23:08:20 nigel Exp $
+ *  $Id: video_macosx.mm,v 1.16 2008/01/01 09:40:32 gbeauche Exp $
  *
  *  video_macosx.mm - Interface between Basilisk II and Cocoa windowing.
  *                    Based on video_amiga.cpp and video_x.cpp
  *
- *  Basilisk II (C) 1997-2003 Christian Bauer
+ *  Basilisk II (C) 1997-2008 Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -101,6 +101,12 @@ colours_from_depth(const uint16 depth)
 bool
 parse_screen_prefs(const char *mode_str)
 {
+	if ( ! mode_str )
+	{
+		// No screen pref was found. Supply a default:
+		mode_str = "win/512/384";
+	}
+
 	if (sscanf(mode_str, "win/%hd/%hd/%hd",
 				&init_width, &init_height, &init_depth) == 3)
 		display_type = DISPLAY_WINDOW;
@@ -1030,6 +1036,7 @@ OSX_monitor::switch_to_current_mode(void)
 			failure = "Could not get base address of screen";
 
 	}
+#ifdef CGIMAGEREF
 	// Clean up the old CGImageRef stuff
 	else if ( display_type == DISPLAY_WINDOW && imageRef )
 	{
@@ -1048,6 +1055,7 @@ OSX_monitor::switch_to_current_mode(void)
 		else
 			failure = "Could not video_open() requested mode";
 	}
+#endif
 	else if ( ! video_open(mode) )
 		failure = "Could not video_open() requested mode";
 
@@ -1103,3 +1111,14 @@ void VideoInterrupt(void)
 void VideoRefresh(void)
 {
 }
+
+
+
+// Deal with a memory access signal referring to the screen.
+// For now, just ignore
+bool Screen_fault_handler(char *a, char *b)
+{
+//	NSLog(@"Got a screen fault %lx %lx", a, b);
+//	[output setNeedsDisplay: YES];
+	return YES;
+}
