@@ -1,7 +1,7 @@
 /*
  *  main_unix.cpp - Startup code for Unix
  *
- *  Basilisk II (C) 1997-2002 Christian Bauer
+ *  Basilisk II (C) Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,7 +24,14 @@
 #include <stdlib.h>
 #include <signal.h>
 #include <errno.h>
-#include <X11/Xlib.h>
+
+#ifdef USE_SDL
+# include <SDL.h>
+#endif
+
+#ifndef USE_SDL_VIDEO
+# include <X11/Xlib.h>
+#endif
 
 #ifdef HAVE_PTHREADS
 # include <pthread.h>
@@ -58,7 +65,7 @@ struct sigstate {
 
 #ifdef ENABLE_XF86_DGA
 # include <X11/Xutil.h>
-# include <X11/extensions/xf86dga.h>
+# include <X11/extensions/Xxf86dga.h>
 #endif
 
 #include <string>
@@ -79,9 +86,10 @@ using std::string;
 #include "main.h"
 #include "vm_alloc.h"
 #include "sigsegv.h"
+#include "rpc.h"
 
 #if USE_JIT
-extern void (*flush_icache)(int); // from compemu_support.cpp
+extern void flush_icache_range(uint8 *start, uint32 size); // from compemu_support.cpp
 #endif
 
 #ifdef ENABLE_MON
@@ -119,8 +127,13 @@ bool TwentyFourBitAddressing;
 
 
 // Global variables
-char *x_display_name = NULL;						// X11 display name
-Display *x_display = NULL;							// X11 display handle
+#ifndef USE_SDL_VIDEO
+extern char *x_display_name;						// X11 display name
+extern Display *x_display;							// X11 display handle
+#ifdef X11_LOCK_TYPE
+X11_LOCK_TYPE x_display_lock = X11_LOCK_INIT;		// X11 display lock
+#endif
+#endif
 
 static uint8 last_xpram[XPRAM_SIZE];				// Buffer for monitoring XPRAM changes
 
@@ -179,6 +192,9 @@ static void sigint_handler(...);
 static bool lm_area_mapped = false;	// Flag: Low Memory area mmap()ped
 #endif
 
+static rpc_connection_t *gui_connection = NULL;	// RPC connection to the GUI
+static const char *gui_connection_path = NULL;	// GUI connection identifier
+
 
 // Prototypes
 static void *xpram_func(void *arg);
@@ -209,14 +225,59 @@ char *strdup(const char *s)
 }
 
 
+/*
+ *  Helpers to map memory that can be accessed from the Mac side
+ */
+
+// NOTE: VM_MAP_32BIT is only used when compiling a 64-bit JIT on specific platforms
+void *vm_acquire_mac(size_t size)
+{
+	return vm_acquire(size, VM_MAP_DEFAULT | VM_MAP_32BIT);
+}
+
+static int vm_acquire_mac_fixed(void *addr, size_t size)
+{
+	return vm_acquire_fixed(addr, size, VM_MAP_DEFAULT | VM_MAP_32BIT);
+}
+
+
+/*
+ *  SIGSEGV handler
+ */
+
+static sigsegv_return_t sigsegv_handler(sigsegv_info_t *sip)
+{
+	const uintptr fault_address = (uintptr)sigsegv_get_fault_address(sip);
+#if ENABLE_VOSF
+	// Handle screen fault
+	extern bool Screen_fault_handler(sigsegv_info_t *sip);
+	if (Screen_fault_handler(sip))
+		return SIGSEGV_RETURN_SUCCESS;
+#endif
+
+#ifdef HAVE_SIGSEGV_SKIP_INSTRUCTION
+	// Ignore writes to ROM
+	if (((uintptr)fault_address - (uintptr)ROMBaseHost) < ROMSize)
+		return SIGSEGV_RETURN_SKIP_INSTRUCTION;
+
+	// Ignore all other faults, if requested
+	if (PrefsFindBool("ignoresegv"))
+		return SIGSEGV_RETURN_SKIP_INSTRUCTION;
+#endif
+
+	return SIGSEGV_RETURN_FAILURE;
+}
+
 /*
  *  Dump state when everything went wrong after a SEGV
  */
 
-static void sigsegv_dump_state(sigsegv_address_t fault_address, sigsegv_address_t fault_instruction)
+static void sigsegv_dump_state(sigsegv_info_t *sip)
 {
+	const sigsegv_address_t fault_address = sigsegv_get_fault_address(sip);
+	const sigsegv_address_t fault_instruction = sigsegv_get_fault_instruction_address(sip);
 	fprintf(stderr, "Caught SIGSEGV at address %p", fault_address);
-	if (fault_instruction != SIGSEGV_INVALID_PC)
+	if (fault_instruction != SIGSEGV_INVALID_ADDRESS)
 		fprintf(stderr, " [IP=%p]", fault_instruction);
 	fprintf(stderr, "\n");
 #if EMULATED_68K
@@ -230,13 +291,75 @@ static void sigsegv_dump_state(sigsegv_address_t fault_address, sigsegv_address_
 #endif
 	VideoQuitFullScreen();
 #ifdef ENABLE_MON
-	char *arg[4] = {"mon", "-m", "-r", NULL};
+	const char *arg[4] = {"mon", "-m", "-r", NULL};
 	mon(3, arg);
-	QuitEmulator();
 #endif
+	QuitEmulator();
 }
 
 
+/*
+ *  Update virtual clock and trigger interrupts if necessary
+ */
+
+#ifdef USE_CPU_EMUL_SERVICES
+static uint64 n_check_ticks = 0;
+static uint64 emulated_ticks_start = 0;
+static uint64 emulated_ticks_count = 0;
+static int64 emulated_ticks_current = 0;
+static int32 emulated_ticks_quantum = 1000;
+int32 emulated_ticks = emulated_ticks_quantum;
+
+void cpu_do_check_ticks(void)
+{
+#if DEBUG
+	n_check_ticks++;
+#endif
+
+	uint64 now;
+	static uint64 next = 0;
+	if (next == 0)
+		next = emulated_ticks_start = GetTicks_usec();
+
+	// Update total instructions count
+	if (emulated_ticks <= 0) {
+		emulated_ticks_current += (emulated_ticks_quantum - emulated_ticks);
+		// XXX: can you really have a machine fast enough to overflow
+		// a 63-bit m68k instruction counter within 16 ms?
+		if (emulated_ticks_current < 0) {
+			printf("WARNING: Overflowed 63-bit m68k instruction counter in less than 16 ms!\n");
+			goto recalibrate_quantum;
+		}
+	}
+
+	// Check for interrupt opportunity
+	now = GetTicks_usec();
+	if (next < now) {
+		one_tick();
+		do {
+			next += 16625;
+		} while (next < now);
+		emulated_ticks_count++;
+
+		// Recalibrate 1000 Hz quantum every 10 ticks
+		static uint64 last = 0;
+		if (last == 0)
+			last = now;
+		else if (now - last > 166250) {
+		  recalibrate_quantum:
+			emulated_ticks_quantum = ((uint64)emulated_ticks_current * 1000) / (now - last);
+			emulated_ticks_current = 0;
+			last = now;
+		}
+	}
+
+	// Update countdown
+	if (emulated_ticks <= 0)
+		emulated_ticks += emulated_ticks_quantum;
+}
+#endif
+
+
 /*
  *  Main program
  */
@@ -251,13 +374,14 @@ static void usage(const char *prg_name)
 		"  --break ADDRESS\n    set ROM breakpoint\n"
 		"  --rominfo\n    dump ROM information\n", prg_name
 	);
-	LoadPrefs(); // read the prefs file so PrefsPrintUsage() will print the correct default values
+	LoadPrefs(NULL); // read the prefs file so PrefsPrintUsage() will print the correct default values
 	PrefsPrintUsage();
 	exit(0);
 }
 
 int main(int argc, char **argv)
 {
+	const char *vmdir = NULL;
 	char str[256];
 
 	// Initialize variables
@@ -274,10 +398,18 @@ int main(int argc, char **argv)
 	for (int i=1; i<argc; i++) {
 		if (strcmp(argv[i], "--help") == 0) {
 			usage(argv[0]);
+#ifndef USE_SDL_VIDEO
 		} else if (strcmp(argv[i], "--display") == 0) {
 			i++; // don't remove the argument, gtk_init() needs it too
 			if (i < argc)
 				x_display_name = strdup(argv[i]);
+#endif
+		} else if (strcmp(argv[i], "--gui-connection") == 0) {
+			argv[i++] = NULL;
+			if (i < argc) {
+				gui_connection_path = argv[i];
+				argv[i] = NULL;
+			}
 		} else if (strcmp(argv[i], "--break") == 0) {
 			argv[i++] = NULL;
 			if (i < argc) {
@@ -311,21 +443,31 @@ int main(int argc, char **argv)
 		}
 	}
 
+	// Connect to the external GUI
+	if (gui_connection_path) {
+		if ((gui_connection = rpc_init_client(gui_connection_path)) == NULL) {
+			fprintf(stderr, "Failed to initialize RPC client connection to the GUI\n");
+			return 1;
+		}
+	}
+
 #ifdef ENABLE_GTK
+	if (!gui_connection) {
 #ifdef HAVE_GNOMEUI
-	// Init GNOME/GTK
-	char version[16];
-	sprintf(version, "%d.%d", VERSION_MAJOR, VERSION_MINOR);
-	gnome_init("Basilisk II", version, argc, argv);
+		// Init GNOME/GTK
+		char version[16];
+		sprintf(version, "%d.%d", VERSION_MAJOR, VERSION_MINOR);
+		gnome_init("Basilisk II", version, argc, argv);
 #else
-	// Init GTK
-	gtk_set_locale();
-	gtk_init(&argc, &argv);
+		// Init GTK
+		gtk_set_locale();
+		gtk_init(&argc, &argv);
 #endif
+	}
 #endif
 
 	// Read preferences
-	PrefsInit(argc, argv);
+	PrefsInit(vmdir, argc, argv);
 
 	// Any command line arguments left?
 	for (int i=1; i<argc; i++) {
@@ -335,6 +477,7 @@ int main(int argc, char **argv)
 		}
 	}
 
+#ifndef USE_SDL_VIDEO
 	// Open display
 	x_display = XOpenDisplay(x_display_name);
 	if (x_display == NULL) {
@@ -348,21 +491,42 @@ int main(int argc, char **argv)
 	// Fork out, so we can return from fullscreen mode when things get ugly
 	XF86DGAForkApp(DefaultScreen(x_display));
 #endif
+#endif
+
+#ifdef USE_SDL
+	// Initialize SDL system
+	int sdl_flags = 0;
+#ifdef USE_SDL_VIDEO
+	sdl_flags |= SDL_INIT_VIDEO;
+#endif
+#ifdef USE_SDL_AUDIO
+	sdl_flags |= SDL_INIT_AUDIO;
+#endif
+	assert(sdl_flags != 0);
+	if (SDL_Init(sdl_flags) == -1) {
+		char str[256];
+		sprintf(str, "Could not initialize SDL: %s.\n", SDL_GetError());
+		ErrorAlert(str);
+		QuitEmulator();
+	}
+	atexit(SDL_Quit);
+#endif
 
 	// Init system routines
 	SysInit();
 
 	// Show preferences editor
-	if (!PrefsFindBool("nogui"))
+	if (!gui_connection && !PrefsFindBool("nogui"))
 		if (!PrefsEditor())
 			QuitEmulator();
 
-	// Register request to ignore all segmentation faults
-#ifdef HAVE_SIGSEGV_SKIP_INSTRUCTION
-	if (PrefsFindBool("ignoresegv"))
-		sigsegv_add_ignore_range(0, ~(0UL), SIGSEGV_TRANSFER_LOAD | SIGSEGV_TRANSFER_STORE);
-#endif
-
+	// Install the handler for SIGSEGV
+	if (!sigsegv_install_handler(sigsegv_handler)) {
+		sprintf(str, GetString(STR_SIG_INSTALL_ERR), "SIGSEGV", strerror(errno));
+		ErrorAlert(str);
+		QuitEmulator();
+	}
+	
 	// Register dump state function when we got mad after a segfault
 	sigsegv_set_dump_state(sigsegv_dump_state);
 
@@ -372,6 +536,8 @@ int main(int argc, char **argv)
 		WarningAlert(GetString(STR_SMALL_RAM_WARN));
 		RAMSize = 1024*1024;
 	}
+	if (RAMSize > 1023*1024*1024)						// Cap to 1023MB (APD crashes at 1GB)
+		RAMSize = 1023*1024*1024;
 
 #if REAL_ADDRESSING || DIRECT_ADDRESSING
 	RAMSize = RAMSize & -getpagesize();					// Round down to page boundary
@@ -383,23 +549,25 @@ int main(int argc, char **argv)
 #if REAL_ADDRESSING
 	// Flag: RAM and ROM are contigously allocated from address 0
 	bool memory_mapped_from_zero = false;
-	
-	// Under Solaris/SPARC and NetBSD/m68k, Basilisk II is known to crash
-	// when trying to map a too big chunk of memory starting at address 0
-#if defined(OS_solaris) || defined(OS_netbsd)
-	const bool can_map_all_memory = false;
-#else
+
+	// Make sure to map RAM & ROM at address 0 only on platforms that
+	// supports linker scripts to relocate the Basilisk II executable
+	// above 0x70000000
+#if HAVE_LINKER_SCRIPT
 	const bool can_map_all_memory = true;
+#else
+	const bool can_map_all_memory = false;
 #endif
 	
 	// Try to allocate all memory from 0x0000, if it is not known to crash
-	if (can_map_all_memory && (vm_acquire_fixed(0, RAMSize + 0x100000) == 0)) {
+	if (can_map_all_memory && (vm_acquire_mac_fixed(0, RAMSize + 0x100000) == 0)) {
 		D(bug("Could allocate RAM and ROM from 0x0000\n"));
 		memory_mapped_from_zero = true;
 	}
 	
+#ifndef PAGEZERO_HACK
 	// Otherwise, just create the Low Memory area (0x0000..0x2000)
-	else if (vm_acquire_fixed(0, 0x2000) == 0) {
+	else if (vm_acquire_mac_fixed(0, 0x2000) == 0) {
 		D(bug("Could allocate the Low Memory globals\n"));
 		lm_area_mapped = true;
 	}
@@ -411,6 +579,7 @@ int main(int argc, char **argv)
 		QuitEmulator();
 	}
 #endif
+#endif /* REAL_ADDRESSING */
 
 	// Create areas for Mac RAM and ROM
 #if REAL_ADDRESSING
@@ -421,17 +590,18 @@ int main(int argc, char **argv)
 	else
 #endif
 	{
-		RAMBaseHost = (uint8 *)vm_acquire(RAMSize);
-		ROMBaseHost = (uint8 *)vm_acquire(0x100000);
-		if (RAMBaseHost == VM_MAP_FAILED || ROMBaseHost == VM_MAP_FAILED) {
+		uint8 *ram_rom_area = (uint8 *)vm_acquire_mac(RAMSize + 0x100000);
+		if (ram_rom_area == VM_MAP_FAILED) {	
 			ErrorAlert(STR_NO_MEM_ERR);
 			QuitEmulator();
 		}
+		RAMBaseHost = ram_rom_area;
+		ROMBaseHost = RAMBaseHost + RAMSize;
 	}
 
 #if USE_SCRATCHMEM_SUBTERFUGE
 	// Allocate scratch memory
-	ScratchMem = (uint8 *)vm_acquire(SCRATCH_MEM_SIZE);
+	ScratchMem = (uint8 *)vm_acquire_mac(SCRATCH_MEM_SIZE);
 	if (ScratchMem == VM_MAP_FAILED) {
 		ErrorAlert(STR_NO_MEM_ERR);
 		QuitEmulator();
@@ -446,8 +616,8 @@ int main(int argc, char **argv)
 	ROMBaseMac = Host2MacAddr(ROMBaseHost);
 #endif
 #if REAL_ADDRESSING
-	RAMBaseMac = (uint32)RAMBaseHost;
-	ROMBaseMac = (uint32)ROMBaseHost;
+	RAMBaseMac = Host2MacAddr(RAMBaseHost);
+	ROMBaseMac = Host2MacAddr(ROMBaseHost);
 #endif
 	D(bug("Mac RAM starts at %p (%08x)\n", RAMBaseHost, RAMBaseMac));
 	D(bug("Mac ROM starts at %p (%08x)\n", ROMBaseHost, ROMBaseMac));
@@ -505,7 +675,7 @@ int main(int argc, char **argv)
 #endif
 
 	// Initialize everything
-	if (!InitAll())
+	if (!InitAll(vmdir))
 		QuitEmulator();
 	D(bug("Initialization complete\n"));
 
@@ -568,6 +738,7 @@ int main(int argc, char **argv)
 	sigaction(SIGINT, &sigint_sa, NULL);
 #endif
 
+#ifndef USE_CPU_EMUL_SERVICES
 #if defined(HAVE_PTHREADS)
 
 	// POSIX threads available, start 60Hz thread
@@ -615,7 +786,9 @@ int main(int argc, char **argv)
 
 	// Start 60Hz timer
 	sigemptyset(&timer_sa.sa_mask);		// Block virtual 68k interrupts during SIGARLM handling
+#if !EMULATED_68K
 	sigaddset(&timer_sa.sa_mask, SIG_IRQ);
+#endif
 	timer_sa.sa_handler = one_tick;
 	timer_sa.sa_flags = SA_ONSTACK | SA_RESTART;
 	if (sigaction(SIGALRM, &timer_sa, NULL) < 0) {
@@ -629,8 +802,9 @@ int main(int argc, char **argv)
 	setitimer(ITIMER_REAL, &req, NULL);
 
 #endif
+#endif
 
-#ifdef HAVE_PTHREADS
+#ifdef USE_PTHREADS_SERVICES
 	// Start XPRAM watchdog thread
 	memcpy(last_xpram, XPRAM, XPRAM_SIZE);
 	xpram_thread_active = (pthread_create(&xpram_thread, NULL, xpram_func, NULL) == 0);
@@ -659,7 +833,13 @@ void QuitEmulator(void)
 	Exit680x0();
 #endif
 
-#if defined(HAVE_PTHREADS)
+#if defined(USE_CPU_EMUL_SERVICES)
+	// Show statistics
+	uint64 emulated_ticks_end = GetTicks_usec();
+	D(bug("%ld ticks in %ld usec = %f ticks/sec [%ld tick checks]\n",
+		  (long)emulated_ticks_count, (long)(emulated_ticks_end - emulated_ticks_start),
+		  emulated_ticks_count * 1000000.0 / (emulated_ticks_end - emulated_ticks_start), (long)n_check_ticks));
+#elif defined(USE_PTHREADS_SERVICES)
 	// Stop 60Hz thread
 	if (tick_thread_active) {
 		tick_thread_cancel = true;
@@ -678,7 +858,7 @@ void QuitEmulator(void)
 	setitimer(ITIMER_REAL, &req, NULL);
 #endif
 
-#ifdef HAVE_PTHREADS
+#ifdef USE_PTHREADS_SERVICES
 	// Stop XPRAM watchdog thread
 	if (xpram_thread_active) {
 		xpram_thread_cancel = true;
@@ -694,11 +874,8 @@ void QuitEmulator(void)
 
 	// Free ROM/RAM areas
 	if (RAMBaseHost != VM_MAP_FAILED) {
-		vm_release(RAMBaseHost, RAMSize);
+		vm_release(RAMBaseHost, RAMSize + 0x100000);
 		RAMBaseHost = NULL;
-	}
-	if (ROMBaseHost != VM_MAP_FAILED) {
-		vm_release(ROMBaseHost, 0x100000);
 		ROMBaseHost = NULL;
 	}
 
@@ -726,8 +903,16 @@ void QuitEmulator(void)
 	PrefsExit();
 
 	// Close X11 server connection
+#ifndef USE_SDL_VIDEO
 	if (x_display)
 		XCloseDisplay(x_display);
+#endif
+
+	// Notify GUI we are about to leave
+	if (gui_connection) {
+		if (rpc_method_invoke(gui_connection, RPC_METHOD_EXIT, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR)
+			rpc_method_wait_for_reply(gui_connection, RPC_TYPE_INVALID);
+	}
 
 	exit(0);
 }
@@ -742,7 +927,7 @@ void FlushCodeCache(void *start, uint32 size)
 {
 #if USE_JIT
     if (UseJIT)
-		flush_icache(-1);
+		flush_icache_range((uint8 *)start, size);
 #endif
 #if !EMULATED_68K && defined(__NetBSD__)
 	m68k_sync_icache(start, size);
@@ -763,7 +948,7 @@ static void sigint_handler(...)
 	m68k_dumpstate(&nextpc);
 #endif
 	VideoQuitFullScreen();
-	char *arg[4] = {"mon", "-m", "-r", NULL};
+	const char *arg[4] = {"mon", "-m", "-r", NULL};
 	mon(3, arg);
 	QuitEmulator();
 }
@@ -815,7 +1000,7 @@ struct B2_mutex {
 	    pthread_mutexattr_init(&attr);
 	    // Initialize the mutex for priority inheritance --
 	    // required for accurate timing.
-#ifdef HAVE_PTHREAD_MUTEXATTR_SETPROTOCOL
+#if defined(HAVE_PTHREAD_MUTEXATTR_SETPROTOCOL) && !defined(__CYGWIN__)
 	    pthread_mutexattr_setprotocol(&attr, PTHREAD_PRIO_INHERIT);
 #endif
 #if defined(HAVE_PTHREAD_MUTEXATTR_SETTYPE) && defined(PTHREAD_MUTEX_NORMAL)
@@ -933,7 +1118,7 @@ static void xpram_watchdog(void)
 	}
 }
 
-#ifdef HAVE_PTHREADS
+#ifdef USE_PTHREADS_SERVICES
 static void *xpram_func(void *arg)
 {
 	while (!xpram_thread_cancel) {
@@ -958,7 +1143,7 @@ static void one_second(void)
 	SetInterruptFlag(INTFLAG_1HZ);
 	TriggerInterrupt();
 
-#ifndef HAVE_PTHREADS
+#ifndef USE_PTHREADS_SERVICES
 	static int second_counter = 0;
 	if (++second_counter > 60) {
 		second_counter = 0;
@@ -975,9 +1160,13 @@ static void one_tick(...)
 		one_second();
 	}
 
-#ifndef HAVE_PTHREADS
-	// No threads available, perform video refresh and networking from here
+#ifndef USE_PTHREADS_SERVICES
+	// Threads not used to trigger interrupts, perform video refresh from here
 	VideoRefresh();
+#endif
+
+#ifndef HAVE_PTHREADS
+	// No threads available, perform networking from here
 	SetInterruptFlag(INTFLAG_ETHER);
 #endif
 
@@ -988,7 +1177,7 @@ static void one_tick(...)
 	}
 }
 
-#ifdef HAVE_PTHREADS
+#ifdef USE_PTHREADS_SERVICES
 static void *tick_func(void *arg)
 {
 	uint64 start = GetTicks_usec();
@@ -1005,7 +1194,7 @@ static void *tick_func(void *arg)
 		ticks++;
 	}
 	uint64 end = GetTicks_usec();
-	D(bug("%Ld ticks in %Ld usec = %f ticks/sec\n", ticks, end - start, ticks * 1000000.0 / (end - start)));
+	D(bug("%lld ticks in %lld usec = %f ticks/sec\n", ticks, end - start, ticks * 1000000.0 / (end - start)));
 	return NULL;
 }
 #endif
@@ -1365,7 +1554,12 @@ void display_alert(int title_id, int prefix_id, int button_id, const char *text)
 
 void ErrorAlert(const char *text)
 {
-#ifdef ENABLE_GTK
+	if (gui_connection) {
+		if (rpc_method_invoke(gui_connection, RPC_METHOD_ERROR_ALERT, RPC_TYPE_STRING, text, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR &&
+			rpc_method_wait_for_reply(gui_connection, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR)
+			return;
+	}
+#if defined(ENABLE_GTK) && !defined(USE_SDL_VIDEO)
 	if (PrefsFindBool("nogui") || x_display == NULL) {
 		printf(GetString(STR_SHELL_ERROR_PREFIX), text);
 		return;
@@ -1384,7 +1578,12 @@ void ErrorAlert(const char *text)
 
 void WarningAlert(const char *text)
 {
-#ifdef ENABLE_GTK
+	if (gui_connection) {
+		if (rpc_method_invoke(gui_connection, RPC_METHOD_WARNING_ALERT, RPC_TYPE_STRING, text, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR &&
+			rpc_method_wait_for_reply(gui_connection, RPC_TYPE_INVALID) == RPC_ERROR_NO_ERROR)
+			return;
+	}
+#if defined(ENABLE_GTK) && !defined(USE_SDL_VIDEO)
 	if (PrefsFindBool("nogui") || x_display == NULL) {
 		printf(GetString(STR_SHELL_WARNING_PREFIX), text);
 		return;
