@@ -1,7 +1,7 @@
 /*
  *  serial_unix.cpp - Serial device driver, Unix specific stuff
  *
- *  Basilisk II (C) 1997-2002 Christian Bauer
+ *  Basilisk II (C) 1997-2008 Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -51,6 +51,17 @@ extern "C" {
 #define MONITOR 0
 
 
+// IRIX missing or unsupported defines
+#ifdef sgi
+#ifndef CRTSCTS
+#define CRTSCTS CNEW_RTSCTS
+#endif
+#ifndef B230400
+#define B230400 B115200
+#endif
+#endif
+
+
 // Missing functions
 #ifndef HAVE_CFMAKERAW
 static int cfmakeraw(struct termios *termios_p)
@@ -576,7 +587,7 @@ bool XSERDPort::open_pty(void)
 		// <should we drop privileges if running setuid?>
 
 		// Let the shell do the dirty work
-		execlp("/bin/sh", "/bin/sh", "-c", ++device_name, 0);
+		execlp("/bin/sh", "/bin/sh", "-c", ++device_name, (char *)NULL);
 
 		// exec failed!
 		printf("serial_open:  could not exec %s: %s\n",
