@@ -1,7 +1,7 @@
 /*
  *  sys_unix.cpp - System dependent routines, Unix implementation
  *
- *  Basilisk II (C) 1997-2003 Christian Bauer
+ *  Basilisk II (C) Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,67 +24,99 @@
 #include <sys/stat.h>
 #include <errno.h>
 
+#ifdef HAVE_AVAILABILITYMACROS_H
+#include <AvailabilityMacros.h>
+#endif
+
 #ifdef __linux__
 #include <sys/mount.h>
 #include <linux/cdrom.h>
 #include <linux/fd.h>
 #include <linux/major.h>
 #include <linux/kdev_t.h>
-#include <linux/unistd.h>
 #include <dirent.h>
-
-#ifdef __NR__llseek
-_syscall5(int, _llseek, unsigned int, fd, unsigned long, hi, unsigned long, lo, loff_t *, res, unsigned int, wh);
-#else
-static int _llseek(unsigned int fd, unsigned long hi, unsigned long lo, loff_t *res, unsigned int wh)
-{
-	if (hi)
-		return -1;
-	*res = lseek(fd, lo, wh);
-	if (*res == -1)
-		return -1;
-	return 0;
-}
-#endif
+#include <limits.h>
 #endif
 
 #if defined(__FreeBSD__) || defined(__NetBSD__)
 #include <sys/cdio.h>
 #endif
 
+#if defined __APPLE__ && defined __MACH__
+#include <sys/disk.h>
+#if (defined AQUA || defined HAVE_FRAMEWORK_COREFOUNDATION)
+#ifndef __MACOSX__
+#define __MACOSX__ MAC_OS_X_VERSION_MIN_REQUIRED
+#endif
+#endif
+#endif
+
 #include "main.h"
 #include "macos_util.h"
 #include "prefs.h"
 #include "user_strings.h"
 #include "sys.h"
 
+#if defined(BINCUE)
+#include "bincue_unix.h"
+#endif
+
+#if defined(HAVE_LIBVHD)
+#include "vhd_unix.h"
+#endif
+
+
 #define DEBUG 0
 #include "debug.h"
 
-
 // File handles are pointers to these structures
-struct file_handle {
-	char *name;		// Copy of device/file name
+struct mac_file_handle {
+	char *name;	        // Copy of device/file name
 	int fd;
+
 	bool is_file;		// Flag: plain file or /dev/something?
 	bool is_floppy;		// Flag: floppy device
 	bool is_cdrom;		// Flag: CD-ROM device
 	bool read_only;		// Copy of Sys_open() flag
+
 	loff_t start_byte;	// Size of file header (if any)
 	loff_t file_size;	// Size of file data (only valid if is_file is true)
 
+	bool is_media_present;		// Flag: media is inserted and available
+
 #if defined(__linux__)
 	int cdrom_cap;		// CD-ROM capability flags (only valid if is_cdrom is true)
 #elif defined(__FreeBSD__)
 	struct ioc_capability cdrom_cap;
 #elif defined(__APPLE__) && defined(__MACH__)
-	char	*ioctl_name;	// For CDs on OS X - a device for special ioctls
-	int		ioctl_fd;
+	char *ioctl_name;	// For CDs on OS X - a device for special ioctls
+	int ioctl_fd;
+#endif
+
+#if defined(BINCUE)
+	bool is_bincue;		// Flag: BIN CUE file
+	void *bincue_fd;
+#endif
+
+#if defined(HAVE_LIBVHD)
+	bool is_vhd;		// Flag: VHD file
+	void *vhd_fd;
 #endif
 };
 
+// Open file handles
+struct open_mac_file_handle {
+	mac_file_handle *fh;
+	open_mac_file_handle *next;
+};
+static open_mac_file_handle *open_mac_file_handles = NULL;
+
 // File handle of first floppy drive (for SysMountFirstFloppy())
-static file_handle *first_floppy = NULL;
+static mac_file_handle *first_floppy = NULL;
+
+// Prototypes
+static void cdrom_close(mac_file_handle *fh);
+static bool cdrom_open(mac_file_handle *fh, const char *path = NULL);
 
 
 /*
@@ -93,6 +125,10 @@ static file_handle *first_floppy = NULL;
 
 void SysInit(void)
 {
+#if defined __MACOSX__
+	extern void DarwinSysInit(void);
+	DarwinSysInit();
+#endif
 }
 
 
@@ -102,6 +138,113 @@ void SysInit(void)
 
 void SysExit(void)
 {
+#if defined __MACOSX__
+	extern void DarwinSysExit(void);
+	DarwinSysExit();
+#endif
+}
+
+
+/*
+ *  Manage open file handles
+ */
+
+static void sys_add_mac_file_handle(mac_file_handle *fh)
+{
+	open_mac_file_handle *p = new open_mac_file_handle;
+	p->fh = fh;
+	p->next = open_mac_file_handles;
+	open_mac_file_handles = p;
+}
+
+static void sys_remove_mac_file_handle(mac_file_handle *fh)
+{
+	open_mac_file_handle *p = open_mac_file_handles;
+	open_mac_file_handle *q = NULL;
+
+	while (p) {
+		if (p->fh == fh) {
+			if (q)
+				q->next = p->next;
+			else
+				open_mac_file_handles = p->next;
+			delete p;
+			break;
+		}
+		q = p;
+		p = p->next;
+	}
+}
+
+
+/*
+ *  Account for media that has just arrived
+ */
+
+void SysMediaArrived(const char *path, int type)
+{
+	// Replace the "cdrom" entry (we are polling, it's unique)
+	if (type == MEDIA_CD && !PrefsFindBool("nocdrom"))
+		PrefsReplaceString("cdrom", path);
+
+	// Wait for media to be available for reading
+	if (open_mac_file_handles) {
+		const int MAX_WAIT = 5;
+		for (int i = 0; i < MAX_WAIT; i++) {
+			if (access(path, R_OK) == 0)
+				break;
+			switch (errno) {
+			case ENOENT: // Unlikely
+			case EACCES: // MacOS X is mounting the media
+				sleep(1);
+				continue;
+			}
+			printf("WARNING: Cannot access %s (%s)\n", path, strerror(errno));
+			return;
+		}
+	}
+
+	for (open_mac_file_handle *p = open_mac_file_handles; p != NULL; p = p->next) {
+		mac_file_handle * const fh = p->fh;
+
+		// Re-open CD-ROM device
+		if (fh->is_cdrom && type == MEDIA_CD) {
+			cdrom_close(fh);
+			if (cdrom_open(fh, path)) {
+				fh->is_media_present = true;
+				MountVolume(fh);
+			}
+		}
+	}
+}
+
+
+/*
+ *  Account for media that has just been removed
+ */
+
+void SysMediaRemoved(const char *path, int type)
+{
+	if ((type & MEDIA_REMOVABLE) != MEDIA_CD)
+		return;
+
+	for (open_mac_file_handle *p = open_mac_file_handles; p != NULL; p = p->next) {
+		mac_file_handle * const fh = p->fh;
+
+		// Mark media as not available
+		if (!fh->is_cdrom || !fh->is_media_present)
+			continue;
+		if (fh->name && strcmp(fh->name, path) == 0) {
+			fh->is_media_present = false;
+			break;
+		}
+#if defined __MACOSX__
+		if (fh->ioctl_name && strcmp(fh->ioctl_name, path) == 0) {
+			fh->is_media_present = false;
+			break;
+		}
+#endif
+	}
 }
 
 
@@ -124,29 +267,34 @@ void SysMountFirstFloppy(void)
 void SysAddFloppyPrefs(void)
 {
 #if defined(__linux__)
-	if (access("/dev/.devfsd", F_OK) < 0) {
-		PrefsAddString("floppy", "/dev/fd0u1440");
-		PrefsAddString("floppy", "/dev/fd1u1440");
-	} else {
-		DIR *fd_dir = opendir("/dev/floppy");
-		if (fd_dir) {
-			struct dirent *floppy_dev;
-			while ((floppy_dev = readdir(fd_dir)) != NULL) {
-				if (strstr(floppy_dev->d_name, "u1440") != NULL) {
-					char fd_dev[20];
-					sprintf(fd_dev, "/dev/floppy/%s", floppy_dev->d_name);
-					PrefsAddString("floppy", fd_dev);
-				}
+	DIR *fd_dir = opendir("/dev/floppy");
+	if (fd_dir) {
+		struct dirent *floppy_dev;
+		while ((floppy_dev = readdir(fd_dir)) != NULL) {
+			if (strstr(floppy_dev->d_name, "u1440") != NULL) {
+				char fd_dev[20];
+				sprintf(fd_dev, "/dev/floppy/%s", floppy_dev->d_name);
+				PrefsAddString("floppy", fd_dev);
 			}
-			closedir(fd_dir);
 		}
+		closedir(fd_dir);
+	} else {
+		PrefsAddString("floppy", "/dev/fd0");
+		PrefsAddString("floppy", "/dev/fd1");
 	}
 #elif defined(__NetBSD__)
 	PrefsAddString("floppy", "/dev/fd0a");
 	PrefsAddString("floppy", "/dev/fd1a");
 #elif defined(__APPLE__) && defined(__MACH__)
-	PrefsAddString("floppy", "/dev/fd/0");
-	PrefsAddString("floppy", "/dev/fd/1");
+  #if defined(AQUA) || defined(HAVE_FRAMEWORK_COREFOUNDATION)
+	extern	void DarwinAddFloppyPrefs(void);
+
+	DarwinAddFloppyPrefs();
+  #else
+	// Until I can convince the other guys that my Darwin code is useful,
+	// we just add something safe (a non-existant device):
+	PrefsAddString("floppy", "/dev/null");
+  #endif
 #else
 	PrefsAddString("floppy", "/dev/fd0");
 	PrefsAddString("floppy", "/dev/fd1");
@@ -217,10 +365,12 @@ void SysAddCDROMPrefs(void)
 			closedir(cd_dir);
 		}
 	}
-#elif defined(__APPLE__) && defined(__MACH__)
-	extern	void DarwinAddCDROMPrefs(void);
-
-	DarwinAddCDROMPrefs();
+#elif defined __MACOSX__
+	// There is no predefined path for CD-ROMs on MacOS X. Rather, we
+	// define a single fake CD-ROM entry for the emulated MacOS.
+	// XXX this means we handle only CD-ROM drive at a time, wherever
+	// the disk is, the latest one is used.
+	PrefsAddString("cdrom", "/dev/poll/cdrom");
 #elif defined(__FreeBSD__) || defined(__NetBSD__)
 	PrefsAddString("cdrom", "/dev/cd0c");
 #endif
@@ -248,10 +398,85 @@ void SysAddSerialPrefs(void)
 	PrefsAddString("seriala", "/dev/tty00");
 	PrefsAddString("serialb", "/dev/tty01");
 #elif defined(__APPLE__) && defined(__MACH__)
-	PrefsAddString("seriala", "/dev/ttys0");
-	PrefsAddString("serialb", "/dev/ttys1");
-//	PrefsAddString("seriala", "/dev/cu.modem");
-//	PrefsAddString("serialb", "/dev/cu.IrDA-IrCOMMch-b");
+  #if defined(AQUA) || defined(HAVE_FRAMEWORK_COREFOUNDATION)
+	extern	void DarwinAddSerialPrefs(void);
+
+	DarwinAddSerialPrefs();
+  #else
+	// Until I can convince the other guys that my Darwin code is useful,
+	// we just add something safe (non-existant devices):
+	PrefsAddString("seriala", "/dev/null");
+	PrefsAddString("serialb", "/dev/null");
+  #endif
+#endif
+}
+
+
+/*
+ *  Open CD-ROM device and initialize internal data
+ */
+
+static bool cdrom_open_1(mac_file_handle *fh)
+{
+#if defined __MACOSX__
+	// In OS X, the device name is OK for sending ioctls to,
+	// but not for reading raw CDROM data from.
+	// (it seems to have extra data padded in)
+	//
+	// So, we keep the already opened file handle,
+	// and open a slightly different file for CDROM data 
+	//
+	fh->ioctl_fd = fh->fd;
+	fh->ioctl_name = fh->name;
+	fh->fd = -1;
+	fh->name = (char *)malloc(strlen(fh->ioctl_name) + 3);
+	if (fh->name) {
+		strcpy(fh->name, fh->ioctl_name);
+		strcat(fh->name, "s1");
+		fh->fd = open(fh->name, O_RDONLY, O_NONBLOCK);
+	}
+	if (fh->ioctl_fd < 0)
+		return false;
+#endif
+	return true;
+}
+
+bool cdrom_open(mac_file_handle *fh, const char *path)
+{
+	if (path)
+		fh->name = strdup(path);
+	fh->fd = open(fh->name, O_RDONLY, O_NONBLOCK);
+	fh->start_byte = 0;
+	if (!cdrom_open_1(fh))
+		return false;
+	return fh->fd >= 0;
+}
+
+
+/*
+ *  Close a CD-ROM device
+ */
+
+void cdrom_close(mac_file_handle *fh)
+{
+
+	if (fh->fd >= 0) {
+		close(fh->fd);
+		fh->fd = -1;
+	}
+	if (fh->name) {
+		free(fh->name);
+		fh->name = NULL;
+	}
+#if defined __MACOSX__
+	if (fh->ioctl_fd >= 0) {
+		close(fh->ioctl_fd);
+		fh->ioctl_fd = -1;
+	}
+	if (fh->ioctl_name) {
+		free(fh->ioctl_name);
+		fh->ioctl_name = NULL;
+	}
 #endif
 }
 
@@ -293,6 +518,19 @@ static bool is_drive_mounted(const char *dev_name, char *mount_name)
 /*
  *  Open file/device, create new file handle (returns NULL on error)
  */
+ 
+static mac_file_handle *open_filehandle(const char *name)
+{
+		mac_file_handle *fh = new mac_file_handle;
+		memset(fh, 0, sizeof(mac_file_handle));
+		fh->name = strdup(name);
+		fh->fd = -1;
+#if defined __MACOSX__
+		fh->ioctl_fd = -1;
+		fh->ioctl_name = NULL;
+#endif
+		return fh;
+}
 
 void *Sys_open(const char *name, bool read_only)
 {
@@ -303,24 +541,25 @@ void *Sys_open(const char *name, bool read_only)
 #else
 	bool is_cdrom = strncmp(name, "/dev/cd", 7) == 0;
 #endif
+	bool is_floppy = strncmp(name, "/dev/fd", 7) == 0;
 
-#if defined(__APPLE__) && defined(__MACH__)
-	//
+	bool is_polled_media = strncmp(name, "/dev/poll/", 10) == 0;
+	if (is_floppy) // Floppy open fails if there's no disk inserted
+		is_polled_media = true;
+
+#if defined __MACOSX__
 	// There is no set filename in /dev which is the cdrom,
 	// so we have to see if it is any of the devices that we found earlier
-	//
-	const char	*cdrom;
-	int			tmp = 0;
-
-	while ( (cdrom = PrefsFindString("cdrom", tmp) ) != NULL )
 	{
-		if ( strcmp(name, cdrom) == 0 )
-		{
-			is_cdrom = 1;
-			read_only = 1;
-			break;
+		int index = 0;
+		const char *str;
+		while ((str = PrefsFindString("cdrom", index++)) != NULL) {
+			if (is_polled_media || strcmp(str, name) == 0) {
+				is_cdrom = true;
+				read_only = true;
+				break;
+			}
 		}
-		++tmp;
 	}
 #endif
 
@@ -345,7 +584,39 @@ void *Sys_open(const char *name, bool read_only)
 	}
 
 	// Open file/device
-#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__)
+
+#if defined(BINCUE)
+	void *binfd = open_bincue(name);
+	if (binfd) {
+		mac_file_handle *fh = open_filehandle(name);
+		D(bug("opening %s as bincue\n", name));
+		fh->bincue_fd = binfd;
+		fh->is_bincue = true;
+		fh->read_only = true;
+		fh->is_media_present = true;
+		sys_add_mac_file_handle(fh);
+		return fh;
+	}
+#endif
+
+
+#if defined(HAVE_LIBVHD)
+	int vhdsize;
+	void *vhdfd = vhd_unix_open(name, &vhdsize, read_only);
+	if (vhdfd) {
+		mac_file_handle *fh = open_filehandle(name);
+		D(bug("opening %s as vnd\n", name));
+		fh->is_vhd = true;
+		fh->vhd_fd = vhdfd; 
+		fh->read_only = read_only;
+		fh->file_size = vhdsize;
+		fh->is_media_present = true;
+		sys_add_mac_file_handle(fh);
+		return fh;
+	}
+#endif
+
+#if defined(__linux__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__MACOSX__)
 	int fd = open(name, (read_only ? O_RDONLY : O_RDWR) | (is_cdrom ? O_NONBLOCK : 0));
 #else
 	int fd = open(name, read_only ? O_RDONLY : O_RDWR);
@@ -355,23 +626,18 @@ void *Sys_open(const char *name, bool read_only)
 		read_only = true;
 		fd = open(name, O_RDONLY);
 	}
-	if (fd >= 0) {
-		file_handle *fh = new file_handle;
-		fh->name = strdup(name);
+	if (fd >= 0 || is_polled_media) {
+		mac_file_handle *fh = open_filehandle(name);
 		fh->fd = fd;
 		fh->is_file = is_file;
 		fh->read_only = read_only;
-		fh->start_byte = 0;
-		fh->is_floppy = false;
-		fh->is_cdrom = false;
+		fh->is_floppy = is_floppy;
+		fh->is_cdrom = is_cdrom;
 		if (fh->is_file) {
+			fh->is_media_present = true;
 			// Detect disk image file layout
 			loff_t size = 0;
-#if defined(__linux__)
-			_llseek(fh->fd, 0, 0, &size, SEEK_END);
-#else
 			size = lseek(fd, 0, SEEK_END);
-#endif
 			uint8 data[256];
 			lseek(fd, 0, SEEK_SET);
 			read(fd, data, 256);
@@ -379,6 +645,7 @@ void *Sys_open(const char *name, bool read_only)
 		} else {
 			struct stat st;
 			if (fstat(fd, &st) == 0) {
+				fh->is_media_present = true;
 				if (S_ISBLK(st.st_mode)) {
 					fh->is_cdrom = is_cdrom;
 #if defined(__linux__)
@@ -389,8 +656,6 @@ void *Sys_open(const char *name, bool read_only)
 						if (fh->cdrom_cap < 0)
 							fh->cdrom_cap = 0;
 					}
-#else
-					fh->cdrom_cap = 0;
 #endif
 #elif defined(__FreeBSD__)
 					fh->is_floppy = ((st.st_rdev >> 16) == 2);
@@ -399,46 +664,24 @@ void *Sys_open(const char *name, bool read_only)
 						if (ioctl(fh->fd, CDIOCCAPABILITY, &fh->cdrom_cap) < 0)
 							memset(&fh->cdrom_cap, 0, sizeof(fh->cdrom_cap));
 					}
-#else
-					fh->cdrom_cap = 0;
 #endif
 #elif defined(__NetBSD__)
 					fh->is_floppy = ((st.st_rdev >> 16) == 2);
 #endif
 				}
-#if defined(__APPLE__) && defined(__MACH__)
-
-				// In OS X, the device name is OK for sending ioctls to,
-				// but not for reading raw CDROM data from.
-				// (it seems to have extra data padded in)
-				//
-				// So, we keep the already opened fiole handle,
-				// and open a slightly different file for CDROM data 
-				//
-				if ( is_cdrom )
-				{
-					fh->ioctl_name	= fh->name;
-					fh->ioctl_fd	= fh->fd;
-
-					fh->name = (char *) malloc(strlen(name) + 2);
-					if ( fh->name )
-					{
-						*fh->name = '\0';
-						strcat(fh->name, name);
-						strcat(fh->name, "s1");
-						fh->fd = open(fh->name, (read_only ? O_RDONLY : O_RDWR));
-						if ( fh->fd < 0 ) {
-							printf("WARNING: Cannot open %s (%s)\n",
-											fh->name, strerror(errno));
-							return NULL;
-						}
-					}
+#if defined __MACOSX__
+				if (is_cdrom) {
+					fh->is_cdrom = true;
+					fh->is_floppy = false;
+					if (cdrom_open_1(fh))
+						fh->is_media_present = true;
 				}
 #endif
 			}
 		}
 		if (fh->is_floppy && first_floppy == NULL)
 			first_floppy = fh;
+		sys_add_mac_file_handle(fh);
 		return fh;
 	} else {
 		printf("WARNING: Cannot open %s (%s)\n", name, strerror(errno));
@@ -453,11 +696,26 @@ void *Sys_open(const char *name, bool read_only)
 
 void Sys_close(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return;
 
-	close(fh->fd);
+	sys_remove_mac_file_handle(fh);
+
+#if defined(HAVE_LIBVHD)
+	if (fh->is_vhd)
+		vhd_unix_close(fh->vhd_fd);
+#endif
+
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		close_bincue(fh->bincue_fd);
+#endif
+
+	if (fh->is_cdrom)
+		cdrom_close(fh);
+	if (fh->fd >= 0)
+		close(fh->fd);
 	if (fh->name)
 		free(fh->name);
 	delete fh;
@@ -471,19 +729,23 @@ void Sys_close(void *arg)
 
 size_t Sys_read(void *arg, void *buffer, loff_t offset, size_t length)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return 0;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return read_bincue(fh->bincue_fd, buffer, offset, length);
+#endif
+
+#if defined(HAVE_LIBVHD)
+	if (fh->is_vhd)
+		return vhd_unix_read(fh->vhd_fd, buffer, offset, length);
+#endif
+
 	// Seek to position
-#if defined(__linux__)
-	loff_t pos = offset + fh->start_byte, res;
-	if (_llseek(fh->fd, pos >> 32, pos, &res, SEEK_SET) < 0)
-		return 0;
-#else
 	if (lseek(fh->fd, offset + fh->start_byte, SEEK_SET) < 0)
 		return 0;
-#endif
 
 	// Read data
 	return read(fh->fd, buffer, length);
@@ -497,19 +759,18 @@ size_t Sys_read(void *arg, void *buffer, loff_t offset, size_t length)
 
 size_t Sys_write(void *arg, void *buffer, loff_t offset, size_t length)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return 0;
 
+#if defined(HAVE_LIBVHD)
+	if (fh->is_vhd)
+		return vhd_unix_write(fh->vhd_fd, buffer, offset, length);
+#endif
+
 	// Seek to position
-#if defined(__linux__)
-	loff_t pos = offset + fh->start_byte, res;
-	if (_llseek(fh->fd, pos >> 32, pos, &res, SEEK_SET) < 0)
-		return 0;
-#else
 	if (lseek(fh->fd, offset + fh->start_byte, SEEK_SET) < 0)
 		return 0;
-#endif
 
 	// Write data
 	return write(fh->fd, buffer, length);
@@ -522,10 +783,20 @@ size_t Sys_write(void *arg, void *buffer, loff_t offset, size_t length)
 
 loff_t SysGetFileSize(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return true;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return size_bincue(fh->bincue_fd);
+#endif 
+
+#if defined(HAVE_LIBVHD)
+	if (fh->is_vhd)
+		return fh->file_size;
+#endif
+
 	if (fh->is_file)
 		return fh->file_size;
 	else {
@@ -535,6 +806,16 @@ loff_t SysGetFileSize(void *arg)
 			return 0;
 		D(bug(" BLKGETSIZE returns %d blocks\n", blocks));
 		return (loff_t)blocks * 512;
+#elif defined __MACOSX__
+		uint32 block_size;
+		if (ioctl(fh->ioctl_fd, DKIOCGETBLOCKSIZE, &block_size) < 0)
+			return 0;
+		D(bug(" DKIOCGETBLOCKSIZE returns %lu bytes\n", (unsigned long)block_size));
+		uint64 block_count;
+		if (ioctl(fh->ioctl_fd, DKIOCGETBLOCKCOUNT, &block_count) < 0)
+			return 0;
+		D(bug(" DKIOCGETBLOCKCOUNT returns %llu blocks\n", (unsigned long long)block_count));
+		return block_count * block_size;
 #else
 		return lseek(fh->fd, 0, SEEK_END) - fh->start_byte;
 #endif
@@ -548,16 +829,18 @@ loff_t SysGetFileSize(void *arg)
 
 void SysEject(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return;
 
 #if defined(__linux__)
 	if (fh->is_floppy) {
-		fsync(fh->fd);
-		ioctl(fh->fd, FDFLUSH);
-		ioctl(fh->fd, FDEJECT);
-		close(fh->fd);	// Close and reopen so the driver will see the media change
+		if (fh->fd >= 0) {
+			fsync(fh->fd);
+			ioctl(fh->fd, FDFLUSH);
+			ioctl(fh->fd, FDEJECT);
+			close(fh->fd);	// Close and reopen so the driver will see the media change
+		}
 		fh->fd = open(fh->name, fh->read_only ? O_RDONLY : O_RDWR);
 	} else if (fh->is_cdrom) {
 		ioctl(fh->fd, CDROMEJECT);
@@ -573,31 +856,26 @@ void SysEject(void *arg)
 		fh->fd = open(fh->name, O_RDONLY | O_NONBLOCK);
 	}
 #elif defined(__APPLE__) && defined(__MACH__)
-	if ( fh->is_cdrom ) {
-
-		// Stolen from IOKit/storage/IOMediaBSDClient.h
-		#define DKIOCEJECT _IO('d', 21)
-
+	if (fh->is_cdrom && fh->is_media_present) {
 		close(fh->fd);
-		if ( ioctl(fh->ioctl_fd, DKIOCEJECT) < 0 )
-		{
-			printf("ioctl(DKIOCEJECT) failed on file %s: %s\n",
-								fh->ioctl_name, strerror(errno));
+		fh->fd = -1;
+		if (ioctl(fh->ioctl_fd, DKIOCEJECT) < 0) {
+			D(bug(" DKIOCEJECT failed on file %s: %s\n",
+				   fh->ioctl_name, strerror(errno)));
 
-			// If we are running OSX, the device may be is busy
-			// due to the Finder having the disk mounted and open,
-			// so we have to use another method.
-
-			// The only problem is that this takes about 5 seconds:
-
-			char	*cmd = (char *) malloc(30+sizeof(fh->name));
-
-			if ( ! cmd )
-				return;
+			// If we are running MacOS X, the device may be in busy
+			// state because the Finder has mounted the disk
 			close(fh->ioctl_fd);
-			sprintf(cmd, "diskutil eject %s 2>&1 >/dev/null", fh->name);
+			fh->ioctl_fd = -1;
+
+			// Try to use "diskutil eject" but it can take up to 5
+			// seconds to complete
+			static const char eject_cmd[] = "/usr/sbin/diskutil eject %s 2>&1 >/dev/null";
+			char *cmd = (char *)alloca(strlen(eject_cmd) + strlen(fh->ioctl_name) + 1);
+			sprintf(cmd, eject_cmd, fh->ioctl_name);
 			system(cmd);
 		}
+		fh->is_media_present = false;
 	}
 #endif
 }
@@ -609,7 +887,7 @@ void SysEject(void *arg)
 
 bool SysFormat(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
@@ -624,15 +902,18 @@ bool SysFormat(void *arg)
 
 bool SysIsReadOnly(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return true;
 
 #if defined(__linux__)
 	if (fh->is_floppy) {
-		struct floppy_drive_struct stat;
-		ioctl(fh->fd, FDGETDRVSTAT, &stat);
-		return !(stat.flags & FD_DISK_WRITABLE);
+		if (fh->fd >= 0) {
+			struct floppy_drive_struct stat;
+			ioctl(fh->fd, FDGETDRVSTAT, &stat);
+			return !(stat.flags & FD_DISK_WRITABLE);
+		} else
+			return true;
 	} else
 #endif
 		return fh->read_only;
@@ -645,10 +926,15 @@ bool SysIsReadOnly(void *arg)
 
 bool SysIsFixedDisk(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return true;
 
+#if defined(HAVE_LIBVHD)
+	if (fh->is_vhd)
+		return true;
+#endif
+
 	if (fh->is_file)
 		return true;
 	else if (fh->is_floppy || fh->is_cdrom)
@@ -664,10 +950,15 @@ bool SysIsFixedDisk(void *arg)
 
 bool SysIsDiskInserted(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(HAVE_LIBVHD)
+	if (fh->is_vhd)
+		return true;
+#endif
+
 	if (fh->is_file) {
 		return true;
 
@@ -707,6 +998,9 @@ bool SysIsDiskInserted(void *arg)
 	} else if (fh->is_cdrom) {
 		struct ioc_toc_header header;
 		return ioctl(fh->fd, CDIOREADTOCHEADER, &header) == 0;
+#elif defined __MACOSX__
+	} else if (fh->is_cdrom || fh->is_floppy) {
+		return fh->is_media_present;
 #endif
 
 	} else
@@ -720,7 +1014,7 @@ bool SysIsDiskInserted(void *arg)
 
 void SysPreventRemoval(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return;
 
@@ -737,7 +1031,7 @@ void SysPreventRemoval(void *arg)
 
 void SysAllowRemoval(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return;
 
@@ -754,11 +1048,17 @@ void SysAllowRemoval(void *arg)
 
 bool SysCDReadTOC(void *arg, uint8 *toc)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return readtoc_bincue(fh->bincue_fd, toc);
+#endif
+
 	if (fh->is_cdrom) {
+
 #if defined(__linux__)
 		uint8 *p = toc + 2;
 
@@ -805,10 +1105,12 @@ bool SysCDReadTOC(void *arg, uint8 *toc)
 		*toc++ = toc_size >> 8;
 		*toc++ = toc_size & 0xff;
 		return true;
-#elif defined(__APPLE__) && defined(__MACH__) && defined(MAC_OS_X_VERSION_10_2)
-		extern	bool	DarwinCDReadTOC(char *name, uint8 *toc);
-
-		return	DarwinCDReadTOC(fh->name, toc);
+#elif defined __MACOSX__ && defined MAC_OS_X_VERSION_10_2
+		if (fh->is_media_present) {
+			extern bool DarwinCDReadTOC(char *name, uint8 *toc);
+			return DarwinCDReadTOC(fh->name, toc);
+		}
+		return false;
 #elif defined(__FreeBSD__)
 		uint8 *p = toc + 2;
 
@@ -893,10 +1195,15 @@ bool SysCDReadTOC(void *arg, uint8 *toc)
 
 bool SysCDGetPosition(void *arg, uint8 *pos)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return GetPosition_bincue(fh->bincue_fd, pos);
+#endif
+
 	if (fh->is_cdrom) {
 #if defined(__linux__)
 		cdrom_subchnl chan;
@@ -958,10 +1265,15 @@ bool SysCDGetPosition(void *arg, uint8 *pos)
 
 bool SysCDPlay(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, uint8 end_m, uint8 end_s, uint8 end_f)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return CDPlay_bincue(fh->bincue_fd, start_m, start_s, start_f, end_m, end_s, end_f);
+#endif
+
 	if (fh->is_cdrom) {
 #if defined(__linux__)
 		cdrom_msf play;
@@ -995,10 +1307,15 @@ bool SysCDPlay(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, uint8 end
 
 bool SysCDPause(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return CDPause_bincue(fh->bincue_fd);
+#endif
+
 	if (fh->is_cdrom) {
 #if defined(__linux__)
 		return ioctl(fh->fd, CDROMPAUSE) == 0;
@@ -1018,10 +1335,16 @@ bool SysCDPause(void *arg)
 
 bool SysCDResume(void *arg)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return CDResume_bincue(fh->bincue_fd);
+#endif
+
+
 	if (fh->is_cdrom) {
 #if defined(__linux__)
 		return ioctl(fh->fd, CDROMRESUME) == 0;
@@ -1041,10 +1364,16 @@ bool SysCDResume(void *arg)
 
 bool SysCDStop(void *arg, uint8 lead_out_m, uint8 lead_out_s, uint8 lead_out_f)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
+#if defined(BINCUE)
+	if (fh->is_bincue)
+		return CDStop_bincue(fh->bincue_fd);
+#endif
+
+
 	if (fh->is_cdrom) {
 #if defined(__linux__)
 		return ioctl(fh->fd, CDROMSTOP) == 0;
@@ -1064,7 +1393,7 @@ bool SysCDStop(void *arg, uint8 lead_out_m, uint8 lead_out_s, uint8 lead_out_f)
 
 bool SysCDScan(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, bool reverse)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return false;
 
@@ -1079,7 +1408,7 @@ bool SysCDScan(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, bool reve
 
 void SysCDSetVolume(void *arg, uint8 left, uint8 right)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return;
 
@@ -1105,7 +1434,7 @@ void SysCDSetVolume(void *arg, uint8 left, uint8 right)
 
 void SysCDGetVolume(void *arg, uint8 &left, uint8 &right)
 {
-	file_handle *fh = (file_handle *)arg;
+	mac_file_handle *fh = (mac_file_handle *)arg;
 	if (!fh)
 		return;
 
