@@ -1,7 +1,7 @@
 /*
  *  cdrom.cpp - CD-ROM driver
  *
- *  Basilisk II (C) 1997-2002 Christian Bauer
+ *  Basilisk II (C) 1997-2008 Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -910,6 +910,17 @@ int16 CDROMStatus(uint32 pb, uint32 dce)
 			}
 			return noErr;
 		}
+
+		case 97: {	// WhoIsThere
+			uint8 drives_present = 0;
+			drive_vec::iterator info, end = drives.end();
+			for (info = drives.begin(); info != end; ++info) {
+				if (info->num <= 6)
+					drives_present |= (1 << info->num);
+			}
+			WriteMacInt8(pb + csParam + 1, drives_present);
+			return noErr;
+		}
 	}
 
 	// Drive valid?
@@ -961,7 +972,7 @@ int16 CDROMStatus(uint32 pb, uint32 dce)
 
 		case 121:		// Get CD features
 			WriteMacInt16(pb + csParam, 0x0200);	// 300 KB/s
-			WriteMacInt16(pb + csParam, 0x0300);	// SCSI-2, stereo
+			WriteMacInt16(pb + csParam + 2, 0x0c00);	// SCSI-2, stereo
 			return noErr;
 
 		default:
