@@ -1,7 +1,7 @@
 /*
  *  clip_dummy.cpp - Clipboard handling, dummy implementation
  *
- *  Basilisk II (C) 1997-2002 Christian Bauer
+ *  Basilisk II (C) 1997-2008 Christian Bauer
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 #include "sysdeps.h"
 
 #include "clip.h"
+#include "macos_util.h"
 
 #define DEBUG 0
 #include "debug.h"
@@ -44,6 +45,16 @@ void ClipExit(void)
 }
 
 
+/*
+ *  Mac application reads clipboard
+ */
+
+void GetScrap(void **handle, uint32 type, int32 offset)
+{
+	D(bug("GetScrap handle %p, type %08x, offset %d\n", handle, type, offset));
+}
+
+
 /*
  *  Mac application wrote to clipboard
  */
@@ -55,7 +66,7 @@ void PutScrap(uint32 type, void *scrap, int32 length)
 		return;
 
 	switch (type) {
-		case 'TEXT':
+		case FOURCC('T','E','X','T'):
 			D(bug(" clipping TEXT\n"));
 			break;
 	}
