@@ -4,10 +4,10 @@
  *  Based on work Copyright 1995, 1996 Bernd Schmidt
  *  Changes for UAE-JIT Copyright 2000 Bernd Meyer
  *
- *  Adaptation for Basilisk II and improvements, copyright 2000-2002
+ *  Adaptation for Basilisk II and improvements, copyright 2000-2005
  *    Gwenole Beauchesne
  *
- *  Basilisk II (C) 1997-2002 Christian Bauer
+ *  Basilisk II (C) 1997-2005 Christian Bauer
  *  
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -1100,7 +1100,7 @@ genflags (flagtypes type, wordsizes size, char *value, char *src, char *dst)
 		      "\tint one=scratchie++;\n"
 		      "\tif (needed_flags&FLAG_Z) {\n"
 		      "\tmov_l_ri(zero,0);\n"
-		      "\tmov_l_ri(one,1);\n"
+		      "\tmov_l_ri(one,-1);\n"
 		      "\tmake_flags_live();\n"
 		      "\tcmov_l_rr(zero,one,5);\n"
 		      "\t}\n");
@@ -1123,7 +1123,7 @@ genflags (flagtypes type, wordsizes size, char *value, char *src, char *dst)
 	    comprintf("\tlive_flags();\n");
 	    comprintf("\tif (needed_flags&FLAG_Z) {\n"
 		      "\tcmov_l_rr(zero,one,5);\n"
-		      "\tsetzflg_l(zero);\n"
+		      "\tset_zero(zero, one);\n" /* No longer need one */
 		      "\tlive_flags();\n"
 		      "\t}\n");
 	    comprintf("\tend_needflags();\n");
@@ -1365,6 +1365,7 @@ gen_opcode (unsigned long int opcode)
 	genamode (curi->dmode, "dstreg", curi->size, "dst", 1, 0);
 	start_brace();
 	comprintf("\tint s=scratchie++;\n"
+		  "\tint tmp=scratchie++;\n"
 		  "\tmov_l_rr(s,src);\n");
 	if (curi->size == sz_byte)
 	    comprintf("\tand_l_ri(s,7);\n");
@@ -1380,6 +1381,7 @@ gen_opcode (unsigned long int opcode)
 	     case i_BCLR: op="btr"; break;
 	     case i_BSET: op="bts"; break;
 	     case i_BTST: op="bt"; need_write=0; break;
+	    default: abort();
 	    }
 	    comprintf("\t%s_l_rr(dst,s);\n"  /* Answer now in C */
 				  "\tsbb_l(s,s);\n" /* s is 0 if bit was 0, -1 otherwise */
@@ -1387,7 +1389,7 @@ gen_opcode (unsigned long int opcode)
 				  "\tdont_care_flags();\n",op);
 		if (!noflags) {
 		  comprintf("\tstart_needflags();\n"
-					"\tsetzflg_l(s);\n"
+					"\tset_zero(s,tmp);\n"
 					"\tlive_flags();\n"
 					"\tend_needflags();\n");
 		}
@@ -1547,9 +1549,9 @@ gen_opcode (unsigned long int opcode)
 	start_brace();
 	comprintf("\tint newad=scratchie++;\n"
 		  "\treadlong(15,newad,scratchie);\n"
-		  "\tmov_l_mr((uae_u32)&regs.pc,newad);\n"
+		  "\tmov_l_mr((uintptr)&regs.pc,newad);\n"
 		  "\tget_n_addr_jmp(newad,PC_P,scratchie);\n"
-		  "\tmov_l_mr((uae_u32)&regs.pc_oldp,PC_P);\n"
+		  "\tmov_l_mr((uintptr)&regs.pc_oldp,PC_P);\n"
 		  "\tm68k_pc_offset=0;\n"
 		  "\tadd_l(15,offs);\n");
 	gen_update_next_handler();
@@ -1578,9 +1580,9 @@ gen_opcode (unsigned long int opcode)
      case i_RTS:
 	comprintf("\tint newad=scratchie++;\n"
 		  "\treadlong(15,newad,scratchie);\n"
-		  "\tmov_l_mr((uae_u32)&regs.pc,newad);\n"
+		  "\tmov_l_mr((uintptr)&regs.pc,newad);\n"
 		  "\tget_n_addr_jmp(newad,PC_P,scratchie);\n"
-		  "\tmov_l_mr((uae_u32)&regs.pc_oldp,PC_P);\n"
+		  "\tmov_l_mr((uintptr)&regs.pc_oldp,PC_P);\n"
 		  "\tm68k_pc_offset=0;\n"
 		  "\tlea_l_brr(15,15,4);\n");
 	gen_update_next_handler();
@@ -1603,18 +1605,18 @@ gen_opcode (unsigned long int opcode)
 		  "\tmov_l_ri(ret,retadd);\n"
 		  "\tsub_l_ri(15,4);\n"
 		  "\twritelong_clobber(15,ret,scratchie);\n");
-	comprintf("\tmov_l_mr((uae_u32)&regs.pc,srca);\n"
+	comprintf("\tmov_l_mr((uintptr)&regs.pc,srca);\n"
 		  "\tget_n_addr_jmp(srca,PC_P,scratchie);\n"
-		  "\tmov_l_mr((uae_u32)&regs.pc_oldp,PC_P);\n"
+		  "\tmov_l_mr((uintptr)&regs.pc_oldp,PC_P);\n"
 		  "\tm68k_pc_offset=0;\n");
 	gen_update_next_handler();
 	break;
      case i_JMP:
 	isjump;
 	genamode (curi->smode, "srcreg", curi->size, "src", 0, 0);
-	comprintf("\tmov_l_mr((uae_u32)&regs.pc,srca);\n"
+	comprintf("\tmov_l_mr((uintptr)&regs.pc,srca);\n"
 		  "\tget_n_addr_jmp(srca,PC_P,scratchie);\n"
-		  "\tmov_l_mr((uae_u32)&regs.pc_oldp,PC_P);\n"
+		  "\tmov_l_mr((uintptr)&regs.pc_oldp,PC_P);\n"
 		  "\tm68k_pc_offset=0;\n");
 	gen_update_next_handler();
 	break;
@@ -1645,8 +1647,8 @@ gen_opcode (unsigned long int opcode)
 	comprintf("\tsub_l_ri(src,m68k_pc_offset-m68k_pc_offset_thisinst-2);\n");
 	/* Leave the following as "add" --- it will allow it to be optimized
 	   away due to src being a constant ;-) */
-	comprintf("\tadd_l_ri(src,(uae_u32)comp_pc_p);\n");  
-	comprintf("\tmov_l_ri(PC_P,(uae_u32)comp_pc_p);\n");
+	comprintf("\tadd_l_ri(src,(uintptr)comp_pc_p);\n");  
+	comprintf("\tmov_l_ri(PC_P,(uintptr)comp_pc_p);\n");
 	/* Now they are both constant. Might as well fold in m68k_pc_offset */
 	comprintf("\tadd_l_ri(src,m68k_pc_offset);\n");
 	comprintf("\tadd_l_ri(PC_P,m68k_pc_offset);\n");
@@ -1719,7 +1721,7 @@ gen_opcode (unsigned long int opcode)
 	 default: abort();  /* Seems this only comes in word flavour */
 	}
 	comprintf("\tsub_l_ri(offs,m68k_pc_offset-m68k_pc_offset_thisinst-2);\n"); 
-	comprintf("\tadd_l_ri(offs,(uae_u32)comp_pc_p);\n"); /* New PC, 
+	comprintf("\tadd_l_ri(offs,(uintptr)comp_pc_p);\n"); /* New PC, 
 								once the 
 								offset_68k is
 								* also added */
@@ -1894,51 +1896,45 @@ gen_opcode (unsigned long int opcode)
 		comprintf("\tint highmask;\n"
 			  "\tint width;\n"
 			  "\tint cdata=scratchie++;\n"
-			  "\tint tmpcnt=scratchie++;\n"
-			  "\tint highshift=scratchie++;\n");
-		comprintf("\tmov_l_rr(tmpcnt,cnt);\n"
-			  "\tand_l_ri(tmpcnt,63);\n"
-			  "\tmov_l_ri(cdata,0);\n"
-			  "\tcmov_l_rr(cdata,data,5);\n");
-		/* cdata is now either data (for shift count!=0) or
-		   0 (for shift count==0) */
-		switch(curi->size) {
-		 case sz_byte: comprintf("\tshra_b_rr(data,cnt);\n"
-					 "\thighmask=0x38;\n"
-					 "\twidth=8;\n"); 
-		 break;
-		 case sz_word: comprintf("\tshra_w_rr(data,cnt);\n"
-					 "\thighmask=0x30;\n"
-					 "\twidth=16;\n"); 
-		 break;
-		 case sz_long: comprintf("\tshra_l_rr(data,cnt);\n"
-					 "\thighmask=0x20;\n"
-					 "\twidth=32;\n"); 
-		 break;
-		 default: abort();
+			  "\tint sdata=scratchie++;\n"
+			  "\tint tmpcnt=scratchie++;\n");
+		comprintf("\tmov_l_rr(sdata,data);\n"
+			  "\tmov_l_rr(cdata,data);\n"
+			  "\tmov_l_rr(tmpcnt,cnt);\n");
+		switch (curi->size) {
+		case sz_byte: comprintf("\tshra_b_ri(sdata,7);\n"); break;
+		case sz_word: comprintf("\tshra_w_ri(sdata,15);\n"); break;
+		case sz_long: comprintf("\tshra_l_ri(sdata,31);\n"); break;
+		default: abort();
 		}
-		comprintf("test_l_ri(cnt,highmask);\n"
-			  "mov_l_ri(highshift,0);\n"
-			  "mov_l_ri(scratchie,width/2);\n"
-			  "cmov_l_rr(highshift,scratchie,5);\n");
-		/* The x86 masks out bits, so we now make sure that things
-		   really get shifted as much as planned */
+		/* sdata is now the MSB propagated to all bits for the
+		   register of specified size */
+		comprintf("\tand_l_ri(tmpcnt,63);\n");
 		switch(curi->size) {
-		 case sz_byte: comprintf("\tshra_b_rr(data,highshift);\n");break;
-		 case sz_word: comprintf("\tshra_w_rr(data,highshift);\n");break;
-		 case sz_long: comprintf("\tshra_l_rr(data,highshift);\n");break;
-		 default: abort();
+		case sz_byte: comprintf("\tshra_b_rr(data,tmpcnt);\n"
+					"\thighmask=0x38;\n");
+		    break;
+		case sz_word: comprintf("\tshra_w_rr(data,tmpcnt);\n"
+					"\thighmask=0x30;\n");
+		    break;
+		case sz_long: comprintf("\tshra_l_rr(data,tmpcnt);\n"
+					"\thighmask=0x20;\n");
+		    break;
 		}
-		/* And again */
-		switch(curi->size) {
-		 case sz_byte: comprintf("\tshra_b_rr(data,highshift);\n");break;
-		 case sz_word: comprintf("\tshra_w_rr(data,highshift);\n");break;
-		 case sz_long: comprintf("\tshra_l_rr(data,highshift);\n");break;
-		 default: abort();
+		comprintf("\ttest_l_ri(tmpcnt,highmask);\n");
+		switch (curi->size) {
+		case sz_byte: comprintf("\tcmov_b_rr(data,sdata,NATIVE_CC_NE);\n"); break;
+		case sz_word: comprintf("\tcmov_w_rr(data,sdata,NATIVE_CC_NE);\n"); break;
+		case sz_long: comprintf("\tcmov_l_rr(data,sdata,NATIVE_CC_NE);\n"); break;
 		}
 		
 		/* Result of shift is now in data. Now we need to determine
 		   the carry by shifting cdata one less */
+		/* NOTE: carry bit is cleared if shift count is zero */
+		comprintf("\tmov_l_ri(scratchie,0);\n"
+			  "\ttest_l_rr(tmpcnt,tmpcnt);\n"
+			  "\tcmov_l_rr(sdata,scratchie,NATIVE_CC_EQ);\n"
+			  "\tforget_about(scratchie);\n");
 		comprintf("\tsub_l_ri(tmpcnt,1);\n");
 		switch(curi->size) {
 		 case sz_byte: comprintf("\tshra_b_rr(cdata,tmpcnt);\n");break;
@@ -1947,10 +1943,16 @@ gen_opcode (unsigned long int opcode)
 		 default: abort();
 		}
 		/* If the shift count was higher than the width, we need
-		   to pick up the sign from data */
-		comprintf("test_l_ri(tmpcnt,highmask);\n"
-			  "cmov_l_rr(cdata,data,5);\n");
-		/* And create the flags */
+		   to pick up the sign from original data (sdata) */
+		/* NOTE: for shift count of zero, the following holds
+		   true and cdata contains 0 so that carry bit is cleared */
+		comprintf("\ttest_l_ri(tmpcnt,highmask);\n"
+			  "\tforget_about(tmpcnt);\n"
+			  "\tcmov_l_rr(cdata,sdata,NATIVE_CC_NE);\n");
+
+		/* And create the flags (preserve X flag if shift count is zero) */
+		comprintf("\ttest_l_ri(cnt,63);\n"
+			  "\tcmov_l_rr(FLAGX,cdata,NATIVE_CC_NE);\n");
 		comprintf("\tstart_needflags();\n");
 		comprintf("\tif (needed_flags & FLAG_ZNV)\n");
 		switch(curi->size) {
@@ -1961,7 +1963,6 @@ gen_opcode (unsigned long int opcode)
 		comprintf("\t bt_l_ri(cdata,0);\n"); /* Set C */
 		comprintf("\t live_flags();\n");
 		comprintf("\t end_needflags();\n");
-		comprintf("\t duplicate_carry();\n");
 		comprintf("if (!(needed_flags & FLAG_CZNV)) dont_care_flags();\n");
 		genastore ("data", curi->dmode, "dstreg", curi->size, "data");
 	    }
@@ -2212,43 +2213,46 @@ gen_opcode (unsigned long int opcode)
 		comprintf("\tmov_l_rr(tmpcnt,cnt);\n"
 			  "\tand_l_ri(tmpcnt,63);\n"
 			  "\tmov_l_ri(cdata,0);\n"
-			  "\tcmov_l_rr(cdata,data,5);\n");
+			  "\tcmov_l_rr(cdata,data,NATIVE_CC_NE);\n");
 		/* cdata is now either data (for shift count!=0) or
 		   0 (for shift count==0) */
 		switch(curi->size) {
-		 case sz_byte: comprintf("\tshrl_b_rr(data,cnt);\n"
+		 case sz_byte: comprintf("\tshrl_b_rr(data,tmpcnt);\n"
 					 "\thighmask=0x38;\n"); 
 		 break;
-		 case sz_word: comprintf("\tshrl_w_rr(data,cnt);\n"
+		 case sz_word: comprintf("\tshrl_w_rr(data,tmpcnt);\n"
 					 "\thighmask=0x30;\n"); 
 		 break;
-		 case sz_long: comprintf("\tshrl_l_rr(data,cnt);\n"
+		 case sz_long: comprintf("\tshrl_l_rr(data,tmpcnt);\n"
 					 "\thighmask=0x20;\n"); 
 		 break;
 		 default: abort();
 		}
-		comprintf("test_l_ri(cnt,highmask);\n"
-			  "mov_l_ri(scratchie,0);\n"
-			  "cmov_l_rr(scratchie,data,4);\n");
-		switch(curi->size) {
-		 case sz_byte: comprintf("\tmov_b_rr(data,scratchie);\n");break;
-		 case sz_word: comprintf("\tmov_w_rr(data,scratchie);\n");break;
-		 case sz_long: comprintf("\tmov_l_rr(data,scratchie);\n");break;
-		 default: abort();
+		comprintf("\ttest_l_ri(tmpcnt,highmask);\n"
+			  "\rmov_l_ri(scratchie,0);\n");
+		if (curi->size == sz_long)
+		    comprintf("\tcmov_l_rr(data,scratchie,NATIVE_CC_NE);\n");
+		else {
+		    comprintf("\tcmov_l_rr(scratchie,data,NATIVE_CC_EQ);\n");
+		    switch(curi->size) {
+		    case sz_byte: comprintf("\tmov_b_rr(data,scratchie);\n");break;
+		    case sz_word: comprintf("\tmov_w_rr(data,scratchie);\n");break;
+		    default: abort();
+		    }
 		}
 		/* Result of shift is now in data. Now we need to determine
 		   the carry by shifting cdata one less */
 		comprintf("\tsub_l_ri(tmpcnt,1);\n");
-		switch(curi->size) {
-		 case sz_byte: comprintf("\tshrl_b_rr(cdata,tmpcnt);\n");break;
-		 case sz_word: comprintf("\tshrl_w_rr(cdata,tmpcnt);\n");break;
-		 case sz_long: comprintf("\tshrl_l_rr(cdata,tmpcnt);\n");break;
-		 default: abort();
-		}
-		comprintf("test_l_ri(tmpcnt,highmask);\n"
-			  "mov_l_ri(scratchie,0);\n"
-			  "cmov_l_rr(cdata,scratchie,5);\n");
-		/* And create the flags */
+		comprintf("\tshrl_l_rr(cdata,tmpcnt);\n");
+		comprintf("\ttest_l_ri(tmpcnt,highmask);\n");
+		comprintf("\tforget_about(tmpcnt);\n");
+		if (curi->size != sz_long) /* scratchie is still live for LSR.L */
+		    comprintf("\tmov_l_ri(scratchie,0);\n");
+		comprintf("\tcmov_l_rr(cdata,scratchie,NATIVE_CC_NE);\n");
+		comprintf("\tforget_about(scratchie);\n");
+		/* And create the flags (preserve X flag if shift count is zero) */
+		comprintf("\ttest_l_ri(cnt,63);\n"
+			  "\tcmov_l_rr(FLAGX,cdata,NATIVE_CC_NE);\n");
 		comprintf("\tstart_needflags();\n");
 		comprintf("\tif (needed_flags & FLAG_ZNV)\n");
 		switch(curi->size) {
@@ -2259,7 +2263,6 @@ gen_opcode (unsigned long int opcode)
 		comprintf("\t bt_l_ri(cdata,0);\n"); /* Set C */
 		comprintf("\t live_flags();\n");
 		comprintf("\t end_needflags();\n");
-		comprintf("\t duplicate_carry();\n");
 		comprintf("if (!(needed_flags & FLAG_CZNV)) dont_care_flags();\n");
 		genastore ("data", curi->dmode, "dstreg", curi->size, "data");
 	    }
@@ -2348,56 +2351,61 @@ gen_opcode (unsigned long int opcode)
 		comprintf("\tmov_l_rr(tmpcnt,cnt);\n"
 			  "\tand_l_ri(tmpcnt,63);\n"
 			  "\tmov_l_ri(cdata,0);\n"
-			  "\tcmov_l_rr(cdata,data,5);\n");
+			  "\tcmov_l_rr(cdata,data,NATIVE_CC_NE);\n");
 		/* cdata is now either data (for shift count!=0) or
 		   0 (for shift count==0) */
 		switch(curi->size) {
-		 case sz_byte: comprintf("\tshll_b_rr(data,cnt);\n"
+		 case sz_byte: comprintf("\tshll_b_rr(data,tmpcnt);\n"
 					 "\thighmask=0x38;\n"); 
 		 break;
-		 case sz_word: comprintf("\tshll_w_rr(data,cnt);\n"
+		 case sz_word: comprintf("\tshll_w_rr(data,tmpcnt);\n"
 					 "\thighmask=0x30;\n"); 
 		 break;
-		 case sz_long: comprintf("\tshll_l_rr(data,cnt);\n"
+		 case sz_long: comprintf("\tshll_l_rr(data,tmpcnt);\n"
 					 "\thighmask=0x20;\n"); 
 		 break;
 		 default: abort();
 		}
-		comprintf("test_l_ri(cnt,highmask);\n"
-			  "mov_l_ri(scratchie,0);\n"
-			  "cmov_l_rr(scratchie,data,4);\n");
-		switch(curi->size) {
-		 case sz_byte: comprintf("\tmov_b_rr(data,scratchie);\n");break;
-		 case sz_word: comprintf("\tmov_w_rr(data,scratchie);\n");break;
-		 case sz_long: comprintf("\tmov_l_rr(data,scratchie);\n");break;
-		 default: abort();
+		comprintf("\ttest_l_ri(tmpcnt,highmask);\n"
+			  "\tmov_l_ri(scratchie,0);\n");
+		if (curi->size == sz_long)
+		    comprintf("\tcmov_l_rr(data,scratchie,NATIVE_CC_NE);\n");
+		else {
+		    comprintf("\tcmov_l_rr(scratchie,data,NATIVE_CC_EQ);\n");
+		    switch(curi->size) {
+		    case sz_byte: comprintf("\tmov_b_rr(data,scratchie);\n");break;
+		    case sz_word: comprintf("\tmov_w_rr(data,scratchie);\n");break;
+		    default: abort();
+		    }
 		}
 		/* Result of shift is now in data. Now we need to determine
 		   the carry by shifting cdata one less */
 		comprintf("\tsub_l_ri(tmpcnt,1);\n");
-		switch(curi->size) {
-		 case sz_byte: comprintf("\tshll_b_rr(cdata,tmpcnt);\n");break;
-		 case sz_word: comprintf("\tshll_w_rr(cdata,tmpcnt);\n");break;
-		 case sz_long: comprintf("\tshll_l_rr(cdata,tmpcnt);\n");break;
-		 default: abort();
+		comprintf("\tshll_l_rr(cdata,tmpcnt);\n");
+		comprintf("\ttest_l_ri(tmpcnt,highmask);\n");
+		comprintf("\tforget_about(tmpcnt);\n");
+		if (curi->size != sz_long) /* scratchie is still live for LSL.L */
+		    comprintf("\tmov_l_ri(scratchie,0);\n");
+		comprintf("\tcmov_l_rr(cdata,scratchie,NATIVE_CC_NE);\n");
+		comprintf("\tforget_about(scratchie);\n");
+		/* And create the flags (preserve X flag if shift count is zero) */
+		switch (curi->size) {
+		case sz_byte: comprintf("\tshrl_l_ri(cdata,7);\n"); break;
+		case sz_word: comprintf("\tshrl_l_ri(cdata,15);\n"); break;
+		case sz_long: comprintf("\tshrl_l_ri(cdata,31);\n"); break;
 		}
-		comprintf("test_l_ri(tmpcnt,highmask);\n"
-			  "mov_l_ri(scratchie,0);\n"
-			  "cmov_l_rr(cdata,scratchie,5);\n");
-		/* And create the flags */
+		comprintf("\ttest_l_ri(cnt,63);\n"
+			  "\tcmov_l_rr(FLAGX,cdata,NATIVE_CC_NE);\n");
 		comprintf("\tstart_needflags();\n");
 		comprintf("\tif (needed_flags & FLAG_ZNV)\n");
 		switch(curi->size) {
-		 case sz_byte: comprintf("\t  test_b_rr(data,data);\n"); 
-		    comprintf("\t bt_l_ri(cdata,7);\n"); break;
-		 case sz_word: comprintf("\t  test_w_rr(data,data);\n"); 
-		    comprintf("\t bt_l_ri(cdata,15);\n"); break;
-		 case sz_long: comprintf("\t  test_l_rr(data,data);\n"); 
-		    comprintf("\t bt_l_ri(cdata,31);\n"); break;
+		case sz_byte: comprintf("\t  test_b_rr(data,data);\n"); break;
+		case sz_word: comprintf("\t  test_w_rr(data,data);\n"); break;
+		case sz_long: comprintf("\t  test_l_rr(data,data);\n"); break;
 		}
+		comprintf("\t bt_l_ri(cdata,0);\n");
 		comprintf("\t live_flags();\n");
 		comprintf("\t end_needflags();\n");
-		comprintf("\t duplicate_carry();\n");
 		comprintf("if (!(needed_flags & FLAG_CZNV)) dont_care_flags();\n");
 		genastore ("data", curi->dmode, "dstreg", curi->size, "data");
 	    }
@@ -2766,9 +2774,9 @@ static int postfix;
 static void 
 generate_one_opcode (int rp, int noflags)
 {
-    int i;
     uae_u16 smsk, dmsk;
-    long int opcode = opcode_map[rp];
+    const long int opcode = opcode_map[rp];
+    const char *opcode_str;
     int aborted=0;
     int have_srcreg=0;
     int have_dstreg=0;
@@ -2777,12 +2785,6 @@ generate_one_opcode (int rp, int noflags)
 	|| table68k[opcode].clev > cpu_level)
 	return;
 
-    for (i = 0; lookuptab[i].name[0]; i++)
-    {
-	if (table68k[opcode].mnemo == lookuptab[i].mnemo)
-	    break;
-    }
-
     if (table68k[opcode].handler != -1)
 	return;
 
@@ -2934,6 +2936,8 @@ generate_one_opcode (int rp, int noflags)
     comprintf("\tuae_u32 m68k_pc_offset_thisinst=m68k_pc_offset;\n");
     comprintf("\tm68k_pc_offset+=2;\n");
 
+    opcode_str = get_instruction_string (opcode);
+
     aborted=gen_opcode (opcode);
     {
 	int flags=0;
@@ -2947,19 +2951,19 @@ generate_one_opcode (int rp, int noflags)
 	comprintf ("}\n");
     
 	if (aborted) {
-	    fprintf (stblfile, "{ NULL, 0x%08x, %ld }, /* %s */\n", flags, opcode, lookuptab[i].name);
+	    fprintf (stblfile, "{ NULL, 0x%08x, %ld }, /* %s */\n", flags, opcode, opcode_str);
 	    com_discard();
 	}
 	else {
 	    if (noflags) {
-		fprintf (stblfile, "{ op_%lx_%d_comp_nf, 0x%08x, %ld }, /* %s */\n", opcode, postfix, flags, opcode, lookuptab[i].name);
+		fprintf (stblfile, "{ op_%lx_%d_comp_nf, 0x%08x, %ld }, /* %s */\n", opcode, postfix, flags, opcode, opcode_str);
 		fprintf (headerfile, "extern compop_func op_%lx_%d_comp_nf;\n", opcode, postfix);
-		printf ("void REGPARAM2 op_%lx_%d_comp_nf(uae_u32 opcode) /* %s */\n{\n", opcode, postfix, lookuptab[i].name);
+		printf ("void REGPARAM2 op_%lx_%d_comp_nf(uae_u32 opcode) /* %s */\n{\n", opcode, postfix, opcode_str);
 	    }
 	    else {
-		fprintf (stblfile, "{ op_%lx_%d_comp_ff, 0x%08x, %ld }, /* %s */\n", opcode, postfix, flags, opcode, lookuptab[i].name);
+		fprintf (stblfile, "{ op_%lx_%d_comp_ff, 0x%08x, %ld }, /* %s */\n", opcode, postfix, flags, opcode, opcode_str);
 		fprintf (headerfile, "extern compop_func op_%lx_%d_comp_ff;\n", opcode, postfix);
-		printf ("void REGPARAM2 op_%lx_%d_comp_ff(uae_u32 opcode) /* %s */\n{\n", opcode, postfix, lookuptab[i].name);
+		printf ("void REGPARAM2 op_%lx_%d_comp_ff(uae_u32 opcode) /* %s */\n{\n", opcode, postfix, opcode_str);
 	    }
 	    com_flush();
 	}
@@ -3067,5 +3071,6 @@ main (int argc, char **argv)
     free (table68k);
 	fclose (stblfile);
 	fclose (headerfile);
+    fflush (stdout);
     return 0;
 }
