@@ -14,6 +14,21 @@
  * take care of this.
  *
  * Copyright 1995, 1996 Bernd Schmidt
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
  */
 
 #include <ctype.h>
@@ -30,6 +45,9 @@
 
 #define BOOL_TYPE "int"
 
+/* Define the minimal 680x0 where NV flags are not affected by xBCD instructions.  */
+#define xBCD_KEEPS_NV_FLAGS 4
+
 static FILE *headerfile;
 static FILE *stblfile;
 
@@ -568,7 +586,7 @@ static void duplicate_carry (void)
 }
 
 typedef enum {
-    flag_logical_noclobber, flag_logical, flag_add, flag_sub, flag_cmp, flag_addx, flag_subx, flag_zn,
+    flag_logical_noclobber, flag_logical, flag_add, flag_sub, flag_cmp, flag_addx, flag_subx, flag_z, flag_zn,
     flag_av, flag_sv
 } flagtypes;
 
@@ -622,6 +640,7 @@ static void genflags_normal (flagtypes type, wordsizes size, char *value, char *
     switch (type) {
      case flag_logical_noclobber:
      case flag_logical:
+     case flag_z:
      case flag_zn:
      case flag_av:
      case flag_sv:
@@ -643,6 +662,7 @@ static void genflags_normal (flagtypes type, wordsizes size, char *value, char *
     switch (type) {
      case flag_logical_noclobber:
      case flag_logical:
+     case flag_z:
      case flag_zn:
 	break;
 
@@ -676,6 +696,9 @@ static void genflags_normal (flagtypes type, wordsizes size, char *value, char *
      case flag_sv:
 	printf ("\tSET_VFLG ((flgs ^ flgo) & (flgn ^ flgo));\n");
 	break;
+     case flag_z:
+	printf ("\tSET_ZFLG (GET_ZFLG & (%s == 0));\n", vstr);
+	break;
      case flag_zn:
 	printf ("\tSET_ZFLG (GET_ZFLG & (%s == 0));\n", vstr);
 	printf ("\tSET_NFLG (%s < 0);\n", vstr);
@@ -921,8 +944,16 @@ static void gen_opcode (unsigned long int opcode)
 	printf ("\tif ((((dst & 0xFF) - (src & 0xFF) - (GET_XFLG ? 1 : 0)) & 0x100) > 0xFF) { newv -= 0x60; }\n");
 	printf ("\tSET_CFLG ((((dst & 0xFF) - (src & 0xFF) - bcd - (GET_XFLG ? 1 : 0)) & 0x300) > 0xFF);\n");
 	duplicate_carry ();
-	genflags (flag_zn, curi->size, "newv", "", "");
-	printf ("\tSET_VFLG ((tmp_newv & 0x80) != 0 && (newv & 0x80) == 0);\n");
+	/* Manual says bits NV are undefined though a real 68040 don't change them */
+	if (cpu_level >= xBCD_KEEPS_NV_FLAGS) {
+	    if (next_cpu_level < xBCD_KEEPS_NV_FLAGS)
+		next_cpu_level = xBCD_KEEPS_NV_FLAGS - 1;
+	    genflags (flag_z, curi->size, "newv", "", "");
+	}
+	else {
+	    genflags (flag_zn, curi->size, "newv", "", "");
+	    printf ("\tSET_VFLG ((tmp_newv & 0x80) != 0 && (newv & 0x80) == 0);\n");
+	}
 	genastore ("newv", curi->dmode, "dstreg", curi->size, "dst");
 	break;
      case i_ADD:
@@ -962,8 +993,16 @@ static void gen_opcode (unsigned long int opcode)
 	printf ("\tif (cflg) newv += 0x60;\n");
 	printf ("\tSET_CFLG (cflg);\n");
 	duplicate_carry ();
-	genflags (flag_zn, curi->size, "newv", "", "");
-	printf ("\tSET_VFLG ((tmp_newv & 0x80) == 0 && (newv & 0x80) != 0);\n");
+	/* Manual says bits NV are undefined though a real 68040 don't change them */
+	if (cpu_level >= xBCD_KEEPS_NV_FLAGS) {
+	    if (next_cpu_level < xBCD_KEEPS_NV_FLAGS)
+		next_cpu_level = xBCD_KEEPS_NV_FLAGS - 1;
+	    genflags (flag_z, curi->size, "newv", "", "");
+	}
+	else {
+	    genflags (flag_zn, curi->size, "newv", "", "");
+	    printf ("\tSET_VFLG ((tmp_newv & 0x80) == 0 && (newv & 0x80) != 0);\n");
+	}
 	genastore ("newv", curi->dmode, "dstreg", curi->size, "dst");
 	break;
      case i_NEG:
@@ -993,7 +1032,15 @@ static void gen_opcode (unsigned long int opcode)
 	printf ("\tif (cflg) newv -= 0x60;\n");
 	printf ("\tSET_CFLG (cflg);\n");
 	duplicate_carry();
-	genflags (flag_zn, curi->size, "newv", "", "");
+	/* Manual says bits NV are undefined though a real 68040 don't change them */
+	if (cpu_level >= xBCD_KEEPS_NV_FLAGS) {
+	    if (next_cpu_level < xBCD_KEEPS_NV_FLAGS)
+		next_cpu_level = xBCD_KEEPS_NV_FLAGS - 1;
+	    genflags (flag_z, curi->size, "newv", "", "");
+	}
+	else {
+	    genflags (flag_zn, curi->size, "newv", "", "");
+	}
 	genastore ("newv", curi->smode, "srcreg", curi->size, "src");
 	break;
      case i_CLR:
@@ -2234,38 +2281,38 @@ static int postfix;
 
 static void generate_one_opcode (int rp)
 {
-    int i;
     uae_u16 smsk, dmsk;
     long int opcode = opcode_map[rp];
+    const char *opcode_str;
 
     if (table68k[opcode].mnemo == i_ILLG
 	|| table68k[opcode].clev > cpu_level)
 	return;
 
-    for (i = 0; lookuptab[i].name[0]; i++) {
-	if (table68k[opcode].mnemo == lookuptab[i].mnemo)
-	    break;
-    }
-
     if (table68k[opcode].handler != -1)
 	return;
 
+    opcode_str = get_instruction_string (opcode);
+
     if (opcode_next_clev[rp] != cpu_level) {
+	if (table68k[opcode].flagdead == 0)
+	/* force to the "ff" variant since the instruction doesn't set at all the condition codes */
+	fprintf (stblfile, "{ CPUFUNC_FF(op_%lx_%d), 0, %ld }, /* %s */\n", opcode, opcode_last_postfix[rp],
+		 opcode, opcode_str);
+	else
 	fprintf (stblfile, "{ CPUFUNC(op_%lx_%d), 0, %ld }, /* %s */\n", opcode, opcode_last_postfix[rp],
-		 opcode, lookuptab[i].name);
+		 opcode, opcode_str);
 	return;
     }
 	
 	if (table68k[opcode].flagdead == 0)
 	/* force to the "ff" variant since the instruction doesn't set at all the condition codes */
-    fprintf (stblfile, "{ CPUFUNC_FF(op_%lx_%d), 0, %ld }, /* %s */\n", opcode, postfix, opcode, lookuptab[i].name);
+    fprintf (stblfile, "{ CPUFUNC_FF(op_%lx_%d), 0, %ld }, /* %s */\n", opcode, postfix, opcode, opcode_str);
 	else
-    fprintf (stblfile, "{ CPUFUNC(op_%lx_%d), 0, %ld }, /* %s */\n", opcode, postfix, opcode, lookuptab[i].name);
+    fprintf (stblfile, "{ CPUFUNC(op_%lx_%d), 0, %ld }, /* %s */\n", opcode, postfix, opcode, opcode_str);
 
     fprintf (headerfile, "extern cpuop_func op_%lx_%d_nf;\n", opcode, postfix);
     fprintf (headerfile, "extern cpuop_func op_%lx_%d_ff;\n", opcode, postfix);
-    printf ("void REGPARAM2 CPUFUNC(op_%lx_%d)(uae_u32 opcode) /* %s */\n{\n", opcode, postfix, lookuptab[i].name);
-	printf ("\tcpuop_begin();\n");
 	
 	/* gb-- The "nf" variant for an instruction that doesn't set the condition
 	   codes at all is the same as the "ff" variant, so we don't need the "nf"
@@ -2274,6 +2321,9 @@ static void generate_one_opcode (int rp)
 	if (table68k[opcode].flagdead == 0)
 	printf ("#ifndef NOFLAGS\n");
 
+	printf ("void REGPARAM2 CPUFUNC(op_%lx_%d)(uae_u32 opcode) /* %s */\n{\n", opcode, postfix, opcode_str);
+	printf ("\tcpuop_begin();\n");
+
     switch (table68k[opcode].stype) {
      case 0: smsk = 7; break;
      case 1: smsk = 255; break;
@@ -2393,10 +2443,10 @@ static void generate_one_opcode (int rp)
     gen_opcode (opcode);
     if (need_endlabel)
 	printf ("%s: ;\n", endlabelstr);
-	if (table68k[opcode].flagdead == 0)
-	printf ("\n#endif\n");
 	printf ("\tcpuop_end();\n");
     printf ("}\n");
+	if (table68k[opcode].flagdead == 0)
+	printf ("\n#endif\n");
     opcode_next_clev[rp] = next_cpu_level;
     opcode_last_postfix[rp] = postfix;
 }
@@ -2457,6 +2507,7 @@ static void generate_func (void)
 
 int main (int argc, char **argv)
 {
+    FILE *out;
     read_table68k ();
     do_merges ();
 
@@ -2470,9 +2521,9 @@ int main (int argc, char **argv)
      * cputbl.h that way), but cpuopti can't cope.  That could be fixed, but
      * I don't dare to touch the 68k version.  */
 
-    headerfile = fopen ("cputbl.h", "wb");
-    stblfile = fopen ("cpustbl.cpp", "wb");
-    freopen ("cpuemu.cpp", "wb", stdout);
+    headerfile = fopen ("cputbl.h", "w");
+    stblfile = fopen ("cpustbl.cpp", "w");
+    out = freopen ("cpuemu.cpp", "w", stdout);
 
     generate_includes (stdout);
     generate_includes (stblfile);
@@ -2480,5 +2531,8 @@ int main (int argc, char **argv)
     generate_func ();
 
     free (table68k);
+    fclose (headerfile);
+    fclose (stblfile);
+    fflush (out);
     return 0;
 }
