@@ -1,10 +1,24 @@
- /*
-  * UAE - The Un*x Amiga Emulator
-  *
-  * MC68000 emulation
-  *
-  * (c) 1995 Bernd Schmidt
-  */
+/*
+ * UAE - The Un*x Amiga Emulator
+ *
+ * MC68000 emulation
+ *
+ * (c) 1995 Bernd Schmidt
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -55,12 +69,14 @@ cpuop_func *cpufunctbl[65536];
 
 #if FLIGHT_RECORDER
 struct rec_step {
+	uae_u32 pc;
+#if FLIGHT_RECORDER >= 2
 	uae_u32 d[8];
 	uae_u32 a[8];
-	uae_u32 pc;
+#endif
 };
 
-const int LOG_SIZE = 8192;
+const int LOG_SIZE = 32768;
 static rec_step log[LOG_SIZE];
 static int log_ptr = -1; // First time initialization
 
@@ -72,10 +88,16 @@ static const char *log_filename(void)
 
 void m68k_record_step(uaecptr pc)
 {
-	for (int i = 0; i < 8; i++) {
-		log[log_ptr].d[i] = m68k_dreg(regs, i);
-		log[log_ptr].a[i] = m68k_areg(regs, i);
+#if FLIGHT_RECORDER >= 2
+	/* XXX: if LSB is set, we are recording from generated code and we
+	   don't support registers recording yet.  */
+	if ((pc & 1) == 0) {
+		for (int i = 0; i < 8; i++) {
+			log[log_ptr].d[i] = m68k_dreg(regs, i);
+			log[log_ptr].a[i] = m68k_areg(regs, i);
+		}
 	}
+#endif
 	log[log_ptr].pc = pc;
 	log_ptr = (log_ptr + 1) % LOG_SIZE;
 }
@@ -87,19 +109,34 @@ static void dump_log(void)
 		return;
 	for (int i = 0; i < LOG_SIZE; i++) {
 		int j = (i + log_ptr) % LOG_SIZE;
-		fprintf(f, "pc %08x\n", log[j].pc);
-		fprintf(f, "d0 %08x d1 %08x d2 %08x d3 %08x\n", log[j].d[0], log[j].d[1], log[j].d[2], log[j].d[3]);
-		fprintf(f, "d4 %08x d5 %08x d6 %08x d7 %08x\n", log[j].d[4], log[j].d[5], log[j].d[6], log[j].d[7]);
-		fprintf(f, "a0 %08x a1 %08x a2 %08x a3 %08x\n", log[j].a[0], log[j].a[1], log[j].a[2], log[j].a[3]);
-		fprintf(f, "a4 %08x a5 %08x a6 %08x a7 %08x\n", log[j].a[4], log[j].a[5], log[j].a[6], log[j].a[7]);
+		uae_u32 pc = log[j].pc & ~1;
+		fprintf(f, "pc %08x", pc);
+#if FLIGHT_RECORDER >= 2
+		fprintf(f, "\n");
+		if ((log[j].pc & 1) == 0) {
+			fprintf(f, "d0 %08x d1 %08x d2 %08x d3 %08x\n", log[j].d[0], log[j].d[1], log[j].d[2], log[j].d[3]);
+			fprintf(f, "d4 %08x d5 %08x d6 %08x d7 %08x\n", log[j].d[4], log[j].d[5], log[j].d[6], log[j].d[7]);
+			fprintf(f, "a0 %08x a1 %08x a2 %08x a3 %08x\n", log[j].a[0], log[j].a[1], log[j].a[2], log[j].a[3]);
+			fprintf(f, "a4 %08x a5 %08x a6 %08x a7 %08x\n", log[j].a[4], log[j].a[5], log[j].a[6], log[j].a[7]);
+		}
+#else
+		fprintf(f, " | ");
+#endif
 #if ENABLE_MON
-		disass_68k(f, log[j].pc);
+		disass_68k(f, pc);
 #endif
 	}
 	fclose(f);
 }
 #endif
 
+#if ENABLE_MON
+static void dump_regs(void)
+{
+	m68k_dumpstate(NULL);
+}
+#endif
+
 #define COUNT_INSTRS 0
 
 #if COUNT_INSTRS
@@ -788,12 +825,33 @@ static void Interrupt(int nr)
 
 static int caar, cacr, tc, itt0, itt1, dtt0, dtt1, mmusr, urp, srp;
 
+static int movec_illg (int regno)
+{
+	switch (CPUType) {
+	case 1:
+		if ((regno & 0x7ff) <= 1)
+			return 0;
+		break;
+	case 2:
+	case 3:
+		if ((regno & 0x7ff) <= 2)
+			return 0;
+		if (regno == 3 || regno == 4)
+			return 0;
+		break;
+	case 4:
+		if ((regno & 0x7ff) <= 7) {
+			if (regno != 0x802)
+				return 0;
+		}
+		break;
+	}
+	return 1;
+}
+
 int m68k_move2c (int regno, uae_u32 *regp)
 {
-  if ((CPUType == 1 && (regno & 0x7FF) > 1)
-	  || (CPUType < 4 && (regno & 0x7FF) > 2)
-	  || (CPUType == 4 && regno == 0x802))
-  {
+  if (movec_illg (regno)) {
 	op_illg (0x4E7B);
 	return 0;
   } else {
@@ -809,10 +867,7 @@ int m68k_move2c (int regno, uae_u32 *regp)
 				flush_icache(1);
 		}
 		else {
-			set_cache_state((cacr&0x8000) || 0);
-			// FIXME: The User Manual claims bit 3 of CACR is undefined
-			if (*regp & 0x08)
-				flush_icache(2);
+			set_cache_state(cacr&0x8000);
 		}
 #endif
 	 break;
@@ -839,9 +894,7 @@ int m68k_move2c (int regno, uae_u32 *regp)
 
 int m68k_movec2 (int regno, uae_u32 *regp)
 {
-    if ((CPUType == 1 && (regno & 0x7FF) > 1)
-		|| (CPUType < 4 && (regno & 0x7FF) > 2)
-		|| (CPUType == 4 && regno == 0x802))
+    if (movec_illg (regno))
     {
 	op_illg (0x4E7A);
 	return 0;
@@ -1152,15 +1205,21 @@ void m68k_reset (void)
     fpu_reset();
 	
 #if FLIGHT_RECORDER
-#if ENABLE_MON
-	if (log_ptr == -1) {
-		// Install "log" command in mon
-		mon_add_command("log", dump_log, "log                      Dump m68k emulation log\n");
-	}
-#endif
 	log_ptr = 0;
 	memset(log, 0, sizeof(log));
 #endif
+
+#if ENABLE_MON
+	static bool first_time = true;
+	if (first_time) {
+		first_time = false;
+		mon_add_command("regs", dump_regs, "regs                    Dump m68k emulator registers\n");
+#if FLIGHT_RECORDER
+		// Install "log" command in mon
+		mon_add_command("log", dump_log, "log                      Dump m68k emulation log\n");
+#endif
+	}
+#endif
 }
 
 void m68k_emulop_return(void)
@@ -1320,6 +1379,7 @@ void m68k_do_execute (void)
 		m68k_record_step(m68k_getpc());
 #endif
 		(*cpufunctbl[opcode])(opcode);
+		cpu_check_ticks();
 		if (SPCFLAGS_TEST(SPCFLAG_ALL_BUT_EXEC_RETURN)) {
 			if (m68k_do_specialties())
 				return;
@@ -1327,17 +1387,6 @@ void m68k_do_execute (void)
 	}
 }
 
-#if USE_JIT && !defined(X86_ASSEMBLY)
-void m68k_compile_execute (void)
-{
-    for (;;) {
-	  if (quit_program)
-		break;
-	  m68k_do_compile_execute();
-    }
-}
-#endif
-
 void m68k_execute (void)
 {
 #if USE_JIT
